"""
This module is for visualizing and interacting with time series data that
can be analyzed using dimesionality reduction techniques such as PCA. It
was originally developed to visualize and classify periodic motion data
during running.
"""

from __future__ import annotations

import numpy as np
import pysampled
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
from typing import Optional, List, Dict, Union, Any

from . import utils
from .core import GenericBrowser


class ComponentBrowser(GenericBrowser):
    def __init__(
        self,
        data: np.ndarray,
        data_transform: np.ndarray,
        labels: Optional[np.ndarray] = None,
        figure_handle: Optional[plt.Figure] = None,
        class_names: Optional[Dict[int, str]] = None,
        desired_class_names: Optional[Dict[int, str]] = None,
        annotation_names: Optional[Dict[int, str]] = None,
    ):
        """
        Initialize the ComponentBrowser.

        Args:
            data (np.ndarray): 2D array with number of signals on dim1, and number of time points on dim2.
            data_transform (np.ndarray): Transformed data, still a 2D array with number of signals x number of components.
                For example, transformed using one of (sklearn.decomposition.PCA, umap.UMAP, sklearn.manifold.TSNE, sklearn.decomposition.FastICA)
            labels (Optional[np.ndarray]): n_signals x 1 array with each entry representing the class of each signal piece.
            figure_handle (Optional[plt.Figure]): Matplotlib figure handle.
            class_names (Optional[Dict[int, str]]): Dictionary of class names.
            desired_class_names (Optional[Dict[int, str]]): Dictionary of desired class names.
            annotation_names (Optional[Dict[int, str]]): Dictionary of annotation names.

        This GUI is meant to be used and extended for
          - 'corrections', where classes are modified / assigned
          - 'annotations', where labels or annotations (separate from a class assignment in the sense that each signal belongs exactly to one class, and a signal my have 0 or more annotations)
        """
        super().__init__(figure_handle)
        self.alpha = {"manual": 0.8, "auto": 0.3}
        self.data = data

        n_components = np.shape(data_transform)[1]

        if labels is None:
            self.labels = np.zeros(self.n_signals, dtype=int)
        else:
            assert len(labels) == self.n_signals
            self.labels = labels
        # make sure all class labels are zero or positive!
        assert np.min(self.labels) >= 0

        class_labels = list(np.unique(self.labels))

        # for detecting keypresses
        self.class_labels_str = [str(x) for x in class_labels]
        self.n_classes = len(class_labels)
        if class_names is None:
            self.class_names = {
                class_label: f"Class_{class_label}" for class_label in class_labels
            }
        else:
            assert set(class_names.keys()) == set(class_labels)
            self.class_names = class_names
        self.classes = [
            ClassLabel(label=label, name=self.class_names[label])
            for label in self.labels
        ]

        if desired_class_names is None:
            desired_class_names = self.class_names
        self.desired_class_names = desired_class_names

        if annotation_names is None:
            annotation_names = {1: "Representative", 2: "Best", 3: "Noisy", 4: "Worst"}
        self.annotation_names = annotation_names
        self.annotation_idx_str = [str(x) for x in self.annotation_names]

        self.cid.append(self.figure.canvas.mpl_connect("pick_event", self.onpick))
        self.cid.append(
            self.figure.canvas.mpl_connect(
                "button_press_event", self.select_signal_piece_dblclick
            )
        )

        n_scatter_plots = int(n_components * (n_components - 1) / 2)
        self.gs = GridSpec(3, max(n_scatter_plots, 4))

        self._data_index = 0
        self.plot_handles = {}
        self.plot_handles["ax_pca"] = {}
        plot_number = 0
        for xc in range(n_components - 1):
            for yc in range(xc + 1, n_components):
                this_ax = self.figure.add_subplot(self.gs[1, plot_number])
                this_ax.set_title(str((xc + 1, yc + 1)))
                self.plot_handles["ax_pca"][plot_number] = this_ax
                self.plot_handles[f"scatter_plot_{xc+1}_{yc+1}"] = this_ax.scatter(
                    data_transform[:, xc],
                    data_transform[:, yc],
                    c=self.colors,
                    alpha=self.alpha["auto"],
                    picker=5,
                )
                (self.plot_handles[f"scatter_highlight_{xc+1}_{yc+1}"],) = this_ax.plot(
                    [], [], "o", color="darkorange"
                )
                plot_number += 1

        self.plot_handles["signal_plots"] = []
        this_ax = self.figure.add_subplot(self.gs[2, 0])
        self.plot_handles["ax_signal_plots"] = this_ax
        for signal_count in range(self.n_signals):
            self.plot_handles["signal_plots"].append(
                this_ax.plot(self.data[signal_count, :])[0]
            )

        self.plot_handles["ax_current_signal"] = self.figure.add_subplot(self.gs[2, 1])
        (self.plot_handles["current_signal"],) = self.plot_handles[
            "ax_current_signal"
        ].plot(list(range(self.n_timepts)), [np.nan] * self.n_timepts)
        self.plot_handles["ax_current_signal"].set_xlim(
            self.plot_handles["ax_signal_plots"].get_xlim()
        )
        self.plot_handles["ax_current_signal"].set_ylim(
            self.plot_handles["ax_signal_plots"].get_ylim()
        )

        self.plot_handles["ax_history_signal"] = self.figure.add_subplot(self.gs[2, 2])

        self.plot_handles["ax_signal_full"] = self.figure.add_subplot(self.gs[0, :])
        self.plot_handles["signal_full"] = []
        time_index = np.r_[: self.n_timepts] / self.n_timepts
        for idx, sig in enumerate(self.data):
            (this_plot_handle,) = self.plot_handles["ax_signal_full"].plot(
                idx + time_index, sig, color=self.colors[idx]
            )  # assumes that there is only one line drawn per signal
            self.plot_handles["signal_full"].append(this_plot_handle)
        (self.plot_handles["signal_selected_piece"],) = self.plot_handles[
            "ax_signal_full"
        ].plot([], [], color="gray", linewidth=2)

        this_ylim = self.plot_handles["ax_signal_full"].get_ylim()
        for x_pos in np.r_[: self.n_signals + 1]:  # separators between signals
            self.plot_handles["ax_signal_full"].plot(
                [x_pos] * 2, this_ylim, "k", linewidth=0.2
            )
        self.memoryslots.disable()

        self._class_info_text = utils.TextView([], self.figure, pos="bottom left")
        self.update_class_info_text()

        self._mode = "correction"  # ('correction' or 'annotation')
        self._mode_text = utils.TextView([], self.figure, pos="center left")
        self.update_mode_text()
        self.add_key_binding("m", self.toggle_mode)

        self._annotation_text = utils.TextView(
            ["", "Annotation list:"]
            + [f"{k}:{v}" for k, v in self.annotation_names.items()],
            self.figure,
            pos="top left",
        )

        self._message = utils.TextView(
            ["Last action : "], self.figure, pos="bottom right"
        )

        self._desired_class_info_text = utils.TextView(
            [], self.figure, pos="bottom center"
        )
        self.update_desired_class_info_text()

        self.add_key_binding("r", self.clear_axes)
        plt.show(block=False)

    @property
    def n_signals(self) -> int:
        """Return the number of signals."""
        return self.data.shape[0]

    @property
    def n_timepts(self) -> int:
        """Return the number of time points."""
        return self.data.shape[-1]

    @property
    def colors(self) -> List[tuple]:
        """Return the colors for each class."""
        return [cl.color for cl in self.classes]

    @property
    def signal(self) -> pysampled.Data:
        """Return the 2D Numpy array as a signal."""
        return pysampled.Data(self.data.flatten(), sr=self.n_timepts)

    def select_signal_piece_dblclick(self, event: Any) -> None:
        """Double click a signal piece in the timecourse view to highlight that point.

        Args:
            event (Any): The mouse event.
        """
        if event.inaxes == self.plot_handles["ax_signal_full"] and event.dblclick:
            if 0 <= int(event.xdata) < self.data.shape[0]:
                self._data_index = int(event.xdata)
                self.update()

    def onpick(self, event: Any) -> None:
        """Single click a projected point.

        Args:
            event (Any): The pick event.
        """
        self.pick_event = event
        self._data_index = np.random.choice(event.ind)
        self.update()

    def update(self) -> None:
        """Update the plot with the current data index."""
        super().update()
        for handle_name, handle in self.plot_handles.items():
            if "scatter_plot_" in handle_name:
                this_data = np.squeeze(handle._offsets[self._data_index].data)
                self.plot_handles[
                    handle_name.replace("_plot_", "_highlight_")
                ].set_data(this_data[0], this_data[1])
        self.plot_handles["ax_history_signal"].plot(self.data[self._data_index, :])
        self.plot_handles["current_signal"].set_ydata(self.data[self._data_index, :])
        self.plot_handles["signal_selected_piece"].set_data(
            np.arange(self.n_timepts) / self.n_timepts + self._data_index,
            self.data[self._data_index, :],
        )
        # self.plot_handles['signal_full'][self._data_index].linewidth = 3
        plt.draw()

    def update_class_info_text(self, draw: bool = True) -> None:
        """Update the class info text.

        Args:
            draw (bool): Whether to draw the plot after updating.
        """
        self._class_info_text.update(
            ["Class list:"] + [f"{k}:{v}" for k, v in self.class_names.items()]
        )
        if draw:
            plt.draw()

    def update_desired_class_info_text(self, draw: bool = True) -> None:
        """Update the desired class info text.

        Args:
            draw (bool): Whether to draw the plot after updating.
        """
        self._desired_class_info_text.update(
            ["Desired class list:"]
            + [f"{k}:{v}" for k, v in self.desired_class_names.items()]
        )
        if draw:
            plt.draw()

    def update_mode_text(self, draw: bool = True) -> None:
        """Update the mode text.

        Args:
            draw (bool): Whether to draw the plot after updating.
        """
        self._mode_text.update([f"mode: {self._mode}"])
        if draw:
            plt.draw()

    def update_message_text(self, text: str, draw: bool = True) -> None:
        """Update the message text.

        Args:
            text (str): The message text.
            draw (bool): Whether to draw the plot after updating.
        """
        self._message.update([text])
        if draw:
            plt.draw()

    def toggle_mode(self, event: Optional[Any] = None) -> None:
        """Toggle between correction and annotation modes.

        Args:
            event (Optional[Any]): The key event.
        """
        self._mode = {"correction": "annotation", "annotation": "correction"}[
            self._mode
        ]
        self.update_mode_text()

    def update_colors(
        self, data_idx: Optional[List[int]] = None, draw: bool = True
    ) -> None:
        """Update the colors of the plot.

        Args:
            data_idx (Optional[List[int]]): List of data indices to update.
            draw (bool): Whether to draw the plot after updating.
        """
        if data_idx is None:
            data_idx = list(range(self.n_signals))
        assert isinstance(data_idx, (list, tuple))
        for this_data_idx in data_idx:
            this_color = self.classes[this_data_idx].color
            self.plot_handles["signal_full"][this_data_idx].set_color(this_color)
            for handle_name, handle in self.plot_handles.items():
                if "scatter_plot_" in handle_name:
                    fc = handle.get_facecolors()
                    fc[this_data_idx, :3] = this_color
                    fc[this_data_idx, -1] = (
                        self.alpha["auto"]
                        if self.classes[this_data_idx].is_auto()
                        else self.alpha["manual"]
                    )
                    handle.set_facecolors(fc)
        if draw:
            plt.draw()

    def update_all(self) -> None:
        """Update all components of the plot."""
        self.update()
        self.update_class_info_text(draw=False)
        self.update_mode_text(draw=False)
        self.update_message_text("Default message", draw=False)
        self.update_colors(draw=False)
        plt.draw()

    def clear_axes(self, event: Optional[Any] = None) -> None:
        """Clear the axes.

        Args:
            event (Optional[Any]): The key event.
        """
        self.plot_handles["ax_history_signal"].clear()
        plt.draw()

    def __call__(self, event: Any) -> None:
        """Handle events.

        Args:
            event (Any): The event.
        """
        super().__call__(event)
        if (
            event.name == "key_press_event"
            and event.inaxes == self.plot_handles["ax_signal_full"]
            and (0 <= int(event.xdata) < self.data.shape[0])
        ):
            this_data_idx = int(event.xdata)
            if self._mode == "correction":
                if event.key in self.class_labels_str:
                    new_label = int(event.key)
                    original_label = self.classes[this_data_idx].original_label
                    if new_label == original_label:
                        self.classes[this_data_idx].set_auto()
                    else:
                        self.classes[this_data_idx].set_manual()
                    self.classes[this_data_idx].label = new_label
                    self.update_colors([this_data_idx])
            elif self._mode == "annotation":
                if event.key in self.annotation_idx_str:
                    this_annotation = self.annotation_names[int(event.key)]
                    if this_annotation not in self.classes[this_data_idx].annotations:
                        self.classes[this_data_idx].annotations.append(this_annotation)
                        self.update_message_text(
                            f"Adding annotation {this_annotation} to signal number {this_data_idx}"
                        )

    def classlabels_to_dict(self) -> Dict[int, Dict[str, Union[int, str, List[str]]]]:
        """Convert class labels to a dictionary.

        Returns:
            Dict[int, Dict[str, Union[int, str, List[str]]]]: Dictionary of class labels.
        """
        fields_to_save = (
            "label",
            "name",
            "assignment_type",
            "annotations",
            "original_label",
        )
        ret = {}
        for class_idx, class_label in enumerate(self.classes):
            ret[class_idx] = {fld: getattr(class_label, fld) for fld in fields_to_save}
        return ret

    def set_classlabels(
        self, classlabels_dict: Dict[int, Dict[str, Union[int, str, List[str]]]]
    ) -> None:
        """Set class labels from a dictionary.

        Args:
            classlabels_dict (Dict[int, Dict[str, Union[int, str, List[str]]]]): Dictionary of class labels.
        """
        assert set(classlabels_dict.keys()) == set(range(self.n_signals))
        self.classes = [
            ClassLabel(**this_label) for this_label in classlabels_dict.values()
        ]


class ClassLabel:
    def __init__(
        self,
        label: int,
        name: Optional[str] = None,
        assignment_type: str = "auto",
        annotations: Optional[List[str]] = None,
        original_label: Optional[int] = None,
    ):
        """
        Initialize a ClassLabel.

        Args:
            label (int): Class label (0 - unclassified, 1 - non-resonant, 2 - resonant, etc.).
            name (Optional[str]): Name of the class.
            assignment_type (str): Class label was assigned automatically ('auto') or manually ('manual').
            annotations (Optional[List[str]]): List of annotations for the class instance.
            original_label (Optional[int]): Original class label.
        """
        assert label >= 0
        self._label = int(label)
        if original_label is None:
            self.original_label = label
        else:
            self.original_label = int(original_label)
        if name is None:
            name = f"Class_{label}"
        self.name = name
        assert assignment_type in ("auto", "manual")
        self.assignment_type = assignment_type

        self.palette = plt.get_cmap("tab20")([np.r_[0:1.5:0.05]])[0][:, :3]
        self._update_colors()

        if annotations is None:
            self.annotations = []
        else:
            assert isinstance(annotations, list)
            self.annotations = annotations

    @property
    def color(self) -> tuple:
        """Return the color of the class."""
        if self.is_auto():
            return self.color_auto
        return self.color_manual

    @property
    def label(self) -> int:
        """Return the class label."""
        return self._label

    @label.setter
    def label(self, val: int) -> None:
        """Set the class label.

        Args:
            val (int): The new class label.
        """
        self._label = int(val)
        self._update_colors()

    def _update_colors(self) -> None:
        """Update the colors based on the class label."""
        if self._label == 0:
            self.color_auto = self.color_manual = (0.0, 0.0, 0.0)  # black
        else:
            self.color_auto = self.palette[(self._label - 1) * 2 + 1]  # lighter
            self.color_manual = self.palette[(self._label - 1) * 2]

    def is_auto(self) -> bool:
        """Return whether the class label was assigned automatically.

        Returns:
            bool: True if the class label was assigned automatically, False otherwise.
        """
        return self.assignment_type == "auto"

    def is_manual(self) -> bool:
        """Return whether the class label was assigned manually.

        Returns:
            bool: True if the class label was assigned manually, False otherwise.
        """
        return self.assignment_type == "manual"

    def set_auto(self) -> None:
        """Set the class label assignment type to automatic."""
        self.assignment_type = "auto"

    def set_manual(self) -> None:
        """Set the class label assignment type to manual."""
        self.assignment_type = "manual"

    def add_annotation(self, annot: str) -> None:
        """Add an annotation to the class instance.

        Args:
            annot (str): The annotation to add.
        """
        self.annotations.append(annot)



import os
import pytest
from unittest.mock import patch
import datanavigator._config
from datanavigator._config import (
    set_clip_folder,
    get_clip_folder,
    set_cache_folder,
    get_cache_folder,
)


@pytest.fixture
def temp_folder(tmp_path):
    """Fixture to create a temporary folder."""
    return str(tmp_path)


def test_set_clip_folder(temp_folder):
    """Test setting the clip folder."""
    current_clip_folder = get_clip_folder()
    set_clip_folder(temp_folder)
    assert get_clip_folder() == temp_folder
    # Reset to default value
    set_clip_folder(current_clip_folder)


def test_set_clip_folder_invalid():
    """Test setting an invalid clip folder."""
    with pytest.raises(ValueError, match="The provided folder path does not exist"):
        set_clip_folder("nonexistent_folder")


def test_set_cache_folder(temp_folder):
    """Test setting the cache folder."""
    current_cache_folder = get_cache_folder()
    set_cache_folder(temp_folder)
    assert get_cache_folder() == temp_folder
    # Reset to default value
    set_cache_folder(current_cache_folder)


def test_set_cache_folder_invalid():
    """Test setting an invalid cache folder."""
    with pytest.raises(ValueError, match="The provided folder path does not exist"):
        set_cache_folder("nonexistent_folder")


def test_default_clip_and_cache_folder():
    """Test default clip and cache folder values."""
    curr_clip_folder = get_clip_folder()
    curr_cache_folder = get_cache_folder()
    with patch.dict(
        os.environ, {"CLIP_FOLDER": "C:\\test_clip", "CACHE_FOLDER": "C:\\test_cache"}
    ):
        from importlib import reload

        reload(datanavigator._config)
        assert get_clip_folder() == "C:\\test_clip"
        assert get_cache_folder() == "C:\\test_cache"
    # Reset to default values
    set_clip_folder(curr_clip_folder)
    set_cache_folder(curr_cache_folder)



import os
import pytest
from pathlib import Path
from unittest.mock import patch

import datanavigator

from tests.conftest import simulate_key_press, simulate_mouse_click


@pytest.fixture(scope="function")
def video_fpath(setup_folders):
    return datanavigator.get_example_video()


def test_setup_folders(setup_folders):
    # setup_folders is a fixture that creates temporary directories for clips and cache
    # check if setup_folders is created correctly
    clip_dir, cache_dir = setup_folders
    assert datanavigator.get_clip_folder() == clip_dir
    assert datanavigator.get_cache_folder() == cache_dir


def test_video_browser_init(video_fpath):
    browser = datanavigator.VideoBrowser(video_fpath)
    assert browser.name == "example_video"
    assert browser.fps == browser.data.get_avg_fps() == 29.97
    assert len(browser.data) == 300

    # Check for searching in the clip folder
    clip_dir = datanavigator.get_clip_folder()
    assert os.path.exists(os.path.join(clip_dir, "example_video.mp4"))
    browser2 = datanavigator.VideoBrowser("example_video.mp4")
    assert browser2.name == "example_video"

    with pytest.raises(AssertionError):
        datanavigator.VideoBrowser("non_existent_video.mp4")


def test_video_browser_init2(video_fpath, mock_figure):
    browser = datanavigator.VideoBrowser(video_fpath, figure_or_ax_handle=mock_figure)
    assert browser.name == "example_video"
    assert browser.fps == browser.data.get_avg_fps() == 29.97
    assert len(browser.data) == 300


def test_video_browser_init3(video_fpath, matplotlib_figure):
    fig, ax = matplotlib_figure
    browser = datanavigator.VideoBrowser(video_fpath, figure_or_ax_handle=ax)
    assert browser.name == "example_video"
    assert browser.fps == browser.data.get_avg_fps() == 29.97
    assert len(browser.data) == 300


def test_video_browser_navigation(video_fpath):
    browser = datanavigator.VideoBrowser(video_fpath)
    assert browser._current_idx == 0
    browser(simulate_key_press(browser.figure, "right"))
    assert browser._current_idx == 1
    browser(simulate_key_press(browser.figure, "shift+up"))
    assert browser._current_idx == 299
    browser(simulate_key_press(browser.figure, "shift+down"))
    assert browser._current_idx == 0
    browser(simulate_key_press(browser.figure, "up"))
    browser(simulate_key_press(browser.figure, "up"))
    assert browser._current_idx == 20
    browser(simulate_key_press(browser.figure, "down"))
    assert browser._current_idx == 10
    browser(simulate_key_press(browser.figure, "shift+right"))
    assert browser._current_idx == 13
    browser(simulate_key_press(browser.figure, "shift+left"))
    browser(simulate_key_press(browser.figure, "shift+left"))
    assert browser._current_idx == 7


def test_video_browser_extract_clip(video_fpath):
    browser = datanavigator.VideoBrowser(video_fpath)
    browser(simulate_key_press(browser.figure, "up"))
    browser(simulate_key_press(browser.figure, "up"))
    assert browser._current_idx == 20
    assert browser.memoryslots._list["1"] is None
    browser(simulate_key_press(browser.figure, "1"))
    assert browser.memoryslots._list["1"] == 20
    browser._current_idx = 75
    browser(simulate_key_press(browser.figure, "2"))
    assert browser.memoryslots._list["2"] == 75
    ret_name = browser.extract_clip()
    assert os.path.exists(ret_name)
    fps = browser.fps
    assert Path(ret_name).stem == f"example_video_s{20/fps:.3f}_e{75/fps:.3f}"
    os.remove(ret_name)

    browser._current_idx = 40
    browser(simulate_key_press(browser.figure, "1"))  # go to 20
    assert browser.memoryslots._list["1"] == 20
    browser(simulate_key_press(browser.figure, "1"))  # clear 20 from memory slot 1
    assert browser.memoryslots._list["1"] is None
    browser._current_idx = 40
    browser(simulate_key_press(browser.figure, "1"))
    assert browser.memoryslots._list["1"] == 40

    browser(simulate_key_press(browser.figure, "2"))  # go to 75
    assert browser.memoryslots._list["2"] == 75
    browser(simulate_key_press(browser.figure, "2"))  # clear 75 from memory slot 2
    assert browser.memoryslots._list["2"] is None
    browser._current_idx = 92
    browser(simulate_key_press(browser.figure, "2"))
    assert browser.memoryslots._list["2"] == 92

    def mock_import(name, *args):
        if name == "ffmpeg":
            raise ModuleNotFoundError("No module named 'ffmpeg'")
        return original_import(name, *args)

    original_import = __import__
    with patch("builtins.__import__", side_effect=mock_import):
        ret_name = browser.extract_clip()
        assert os.path.exists(ret_name)
        assert Path(ret_name).stem == f"example_video_s{40/fps:.3f}_e{92/fps:.3f}"
        os.remove(ret_name)


def test_video_plot_browser_init(video_fpath, signal_list):
    # generate random signals for 10 s
    signals = dict(zip(["white_noise", "sine_wave", "three_sine_waves"], signal_list))
    browser = datanavigator.VideoPlotBrowser(video_fpath, signals)
    assert browser.fps == browser.video_data.get_avg_fps() == 29.97
    assert len(browser.video_data) == len(browser) == browser._len == 300

    # test _setup
    assert "signal0" in browser.plot_handles
    assert "signal2_tick" in browser.plot_handles
    # montage is the video
    assert "montage" in browser.plot_handles

    # test mouse click
    assert browser._current_idx == 0
    event = simulate_mouse_click(
        (browser.figure, browser.plot_handles["signal_ax"][0]),
        xdata=0.5,
        ydata=0.5,
        button=3,
    )
    browser.figure.canvas.callbacks.process("button_press_event", event)

    assert browser._current_idx == round(browser.fps * 0.5)

    # test event window and clip extraction
    browser = datanavigator.VideoPlotBrowser(
        video_fpath, signals, event_win=[-0.5, 1.0]
    )
    browser._current_idx = 150
    assert browser.memoryslots._list["1"] == None
    browser(simulate_key_press(browser.figure, "1"))  # go to 20
    assert browser.memoryslots._list["1"] == 150

    browser._current_idx = 160
    assert browser.memoryslots._list["2"] == None
    browser(simulate_key_press(browser.figure, "2"))  # go to 20
    assert browser.memoryslots._list["2"] == 160

    screengrab_clip = browser.extract_clip()
    assert os.path.exists(screengrab_clip)



import pytest
import os
import numpy as np

from datanavigator import events
from datanavigator.events import (
    portion,
    EventData,
    Event,
    Events,
    _find_nearest_idx_val,
)

from tests.conftest import simulate_mouse_click


@pytest.fixture(scope="module")
def event_file_path(tmp_path_factory):
    return tmp_path_factory.mktemp("data") / "picked_event.json"


def test_portion():
    p = portion.closed(1, 2) | portion.closed(3, 5.4)
    assert np.allclose(p.atomic_durations, [1, 2.4])
    assert np.allclose(p.duration, 3.4)
    assert p.enclosure == portion.closed(1, 5.4)
    assert np.allclose(p.fraction, (p.duration / p.enclosure.duration))


def test_event_data_initialization():
    event_data = EventData()
    assert event_data.default == []
    assert event_data.added == []
    assert event_data.removed == []
    assert event_data.tags == []
    assert event_data.algorithm_name == ""
    assert event_data.params == {}

    event_data = EventData(default=[1, 2, 4, 3])
    # list of float/ints are treated as a series of 1-events
    assert event_data.default == [[1], [2], [4], [3]]
    event_data = EventData(default=[1, 2, 4, 3], added=[[5]])
    assert event_data.get_times() == [[1], [2], [3], [4], [5]]
    with pytest.raises(AssertionError, match="All events must have the same size"):
        EventData(default=[1, 2], added=[[3, 4]])
    with pytest.raises(AssertionError, match="All events must have the same size"):
        EventData(default=[1, 2, 3, 4], added=[[3, 4]])


def test_event_data_get_size():
    event_data = EventData(default=[1, 2, 4, 3])
    assert event_data.get_size() == 1
    event_data = EventData(default=[[1, 2], [3, 4]])  # 2 events
    assert event_data.get_size() == 2
    event_data.added.append([5, 6, 7])  # 3 events
    with pytest.raises(ValueError, match="All events must have the same size"):
        event_data.get_size()


def test_event_data_asdict():
    event_data = EventData(
        default=[1],
        added=[2],
        removed=[3],
        tags=["tag"],
        algorithm_name="algo",
        params={"param": "value"},
    )
    expected_dict = {
        "default": [[1]],
        "added": [[2]],
        "removed": [[3]],
        "tags": ["tag"],
        "algorithm_name": "algo",
        "params": {"param": "value"},
    }
    assert event_data.asdict() == expected_dict


def test_event_data_len():
    event_data = EventData(default=[[1, 2]], added=[[3, 4]])
    assert len(event_data) == 2


def test_event_data_get_times():
    event_data = EventData(default=[[3, 4]], added=[[1, 2]])
    assert event_data.get_times() == [[1, 2], [3, 4]]


def test_event_data_to_portions():
    event_data = EventData(default=[[1, 2], [3, 4]])
    expected_portion = portion.closed(1, 2) | portion.closed(3, 4)
    assert event_data.to_portions() == expected_portion


def test_event_data_and():
    event_data1 = EventData(default=[[1, 2], [3, 4]])
    event_data2 = EventData(default=[[2.5, 3.5], [4, 5]])
    expected_portion = portion.closed(3, 3.5)
    # Strictly speaking, since I'm using closed intervals, [4] should also be part of the intersection
    # But I'm only using portions for representing intervals, so I'm not too concerned about this
    assert (event_data1 & event_data2).to_portions() == expected_portion

    event_data1 = EventData(default=[[1, 2], [3, 4]])
    event_data2 = EventData(default=[1, 2, 3, 4])
    with pytest.raises(TypeError):
        event_data1 & event_data2
    event_data1 = EventData(default=[2, 3, 4, 5])
    with pytest.raises(TypeError):
        event_data1 & event_data2


def test_event_data_contains():
    event_data = EventData(default=[[1, 2], [3, 4]])
    assert 1.5 in event_data
    assert 2.5 not in event_data


def test_event_data_overlap():
    event_data1 = EventData([[1, 2], [5.5, 12]])
    event_data2 = EventData([[2, 3], [4, 6]])
    assert np.allclose(event_data1.overlap_duration(event_data2), 0.5)
    assert np.allclose(event_data1.overlap_duration([4, 6]), 0.5)
    assert np.allclose(event_data1.overlap_duration(portion.closed(4, 6)), 0.5)


def test_event_initialization():
    data_id_func = lambda: "test_id"
    event = Event(
        name="test_event", size=2, fname="test.json", data_id_func=data_id_func
    )
    assert event.name == "test_event"
    assert event.size == 2
    assert event.fname == "test.json"
    assert event.data_id_func == data_id_func
    event = Event(
        name="test_event", size=2, fname="test.json", data_id_func=data_id_func, color=3
    )
    assert event.color == events.PLOT_COLORS[3]


def test_event_add_remove(matplotlib_figure, event_file_path):
    data_id_func = lambda: "test_id"
    event = Event(
        name="test_event",
        size=2,
        pick_action="overwrite",
        fname=event_file_path,
        data_id_func=data_id_func,
    )

    # add an event
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.5))
    assert np.allclose(event._buffer, [0.5])
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.7))
    assert event._buffer == []
    assert np.allclose(event._data["test_id"].get_times(), [[0.5, 0.7]])

    # this event should not be added because end > start
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.6))
    assert np.allclose(event._data["test_id"].get_times(), [[0.5, 0.7]])

    # this should overwrite the first event
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.85))
    assert np.allclose(event._data["test_id"].get_times(), [[0.8, 0.85]])

    # add an event when there is one default event
    event = Event(
        name="test_event",
        size=2,
        pick_action="overwrite",
        fname=event_file_path,
        data_id_func=data_id_func,
    )
    event._data["test_id"] = EventData(default=[[0.2, 0.25]])
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.85))
    assert np.allclose(event._data["test_id"].get_times(), [[0.8, 0.85]])
    assert np.allclose(event._data["test_id"].removed, [[0.2, 0.25]])

    ## simulate an event click outside axis limits
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.9))
    event.add(
        simulate_mouse_click(matplotlib_figure, xdata=-0.1)
    )  # outside the axis limits
    assert np.allclose(event._data["test_id"].added, [[0.9, 1.0]])
    event.add(
        simulate_mouse_click(matplotlib_figure, xdata=2.2)
    )  # outside the axis limits
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.22))
    assert np.allclose(event._data["test_id"].added, [[0.0, 0.22]])

    event = Event(
        name="test_event",
        size=2,
        pick_action="append",
        fname=event_file_path,
        data_id_func=data_id_func,
    )
    # add an event when pick_action is append
    event._data["test_id"] = EventData(default=[[0.2, 0.25]])
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.85))
    assert np.allclose(event.to_dict()["test_id"], [[0.2, 0.25], [0.8, 0.85]])

    # delete an event from added, as opposed to removing it
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    assert event._data["test_id"].removed == []

    # remove an event - default events are stored in a "removed" list in the EventData object to keep track of what was removed
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.2))
    assert event._data["test_id"].default == []
    assert event._data["test_id"].added == []
    assert np.allclose(event._data["test_id"].removed, [[0.2, 0.25]])
    assert event.to_dict() == {"test_id": []}

    ## cannot remove an event by clicking outside axis limits
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.85))
    original_dict = event.to_dict()
    event.remove(
        simulate_mouse_click(matplotlib_figure, xdata=-0.1)
    )  # outside the axis limits
    assert event.to_dict() == original_dict
    event.add(
        simulate_mouse_click(matplotlib_figure, xdata=2.2)
    )  # outside the axis limits
    assert event.to_dict() == original_dict

    ## this won't happen unless the user is trying to break the system
    event.data_id_func = lambda: "test_id_2"
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    assert event.to_dict() == original_dict
    event.data_id_func = lambda: "test_id"
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    assert event.to_dict() == {"test_id": []}

    ## removing an event when there are no stored events
    assert event._data["test_id"].default == []
    assert event._data["test_id"].added == []
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    assert event.to_dict() == {"test_id": []}


def test_event_get_current_event_times(event_file_path):
    event = Event(
        name="test_event", size=2, fname=event_file_path, pick_action="overwrite"
    )
    event._data["test_id"] = EventData(default=[[0.2, 0.25], [0.4, 0.45]])
    # with the default data_id_func, get_current_event_times will return events with the key None
    assert event.get_current_event_times() == EventData().get_times()
    event.data_id_func = lambda: "test_id"
    assert np.allclose(event.get_current_event_times(), [[0.2, 0.25], [0.4, 0.45]])


def test_event_to_dict():
    data_id_func = lambda: "test_id"
    event = Event(
        name="test_event", size=2, fname="test.json", data_id_func=data_id_func
    )
    assert event.to_dict() == {}  # test empty event
    event._data["test_id"] = EventData(default=[[1, 2]])
    assert event.to_dict() == {"test_id": [[1, 2]]}

    event = Event(
        name="test_event",
        size=2,
        pick_action="append",
        fname="test.json",
        data_id_func=data_id_func,
    )
    event._data["test_id"] = EventData(default=[[1, 2]])
    assert event.to_dict() == {"test_id": [[1, 2]]}


def test_event_save(matplotlib_figure, event_file_path):
    # add an event when pick_action is append
    data_id_func = lambda: "test_id"
    event = Event(
        name="test_event",
        size=2,
        pick_action="append",
        fname=event_file_path,
        data_id_func=data_id_func,
    )
    event._data["test_id"] = EventData(default=[[0.2, 0.25], [0.1, 0.17], [0.45, 0.55]])
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.8))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.85))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.4))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.6))
    event.remove(simulate_mouse_click(matplotlib_figure, xdata=0.45))
    event.save()


def test_event_from_file(tmp_path_factory, event_file_path):
    assert os.path.exists(event_file_path)
    event = Event.from_file(fname=event_file_path)
    # to_dict will sort the event entries
    assert np.allclose(
        event.to_dict()["test_id"], [[0.1, 0.17], [0.2, 0.25], [0.4, 0.6], [0.8, 0.85]]
    )
    # events in added, default, and removed will remain in the sequence they were added
    assert np.allclose(event._data["test_id"].default, [[0.2, 0.25], [0.1, 0.17]])
    assert np.allclose(event._data["test_id"].added, [[0.8, 0.85], [0.4, 0.6]])
    assert np.allclose(event._data["test_id"].removed, [[0.45, 0.55]])

    fname2 = tmp_path_factory.getbasetemp() / "file_does_not_exist.json"
    assert not os.path.exists(fname2)
    event = Event.from_file(fname2)
    assert os.path.exists(fname2)
    os.remove(fname2)


def test_event_from_data(tmp_path_factory, matplotlib_figure):
    data = {
        (1, 1): [[0.1, 0.2], [0.3, 0.4]],
        (1, 2): [[0.5, 0.6]],
        (2, 1): [[0.7, 0.8], [0.9, 1.0], [0.85, 0.87]],
        (2, 2): [],
    }
    event = Event.from_data(data, name="test_event")
    # only the last events are kept because pick_action is overwrite by default
    assert np.allclose(event._data[(2, 1)].default, [[0.85, 0.87]])
    assert np.allclose(
        event.to_dict()[(2, 1)], [[0.85, 0.87]]
    )  # because pick_action is overwrite

    # the data key is usually handled by the parent UI.
    # This is a function that returns the ID of the plot that the user is current interacting with
    # Since adding an event is a user interaction, it is necessary for adding an event
    event.data_id_func = lambda: (1, 2)
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.75))
    event.add(simulate_mouse_click(matplotlib_figure, xdata=0.82))
    assert np.allclose(event._data[(1, 2)].default, [])
    assert np.allclose(event._data[(1, 2)].removed, [[0.5, 0.6]])
    assert np.allclose(event._data[(1, 2)].added, [[0.75, 0.82]])
    # added events will be treated as being added after default
    assert np.allclose(event.to_dict()[(1, 2)], [[0.75, 0.82]])

    event = Event.from_data(data, name="test_event", pick_action="append")
    # to_dict will sort the event entries
    assert np.allclose(event.to_dict()[(2, 1)], [[0.7, 0.8], [0.85, 0.87], [0.9, 1.0]])

    # just make sure this runs
    data = event._data  # dict[tuple, EventData]
    event = Event.from_data(data, name="test_event")

    fname2 = tmp_path_factory.getbasetemp() / "file_does_not_exist_2.json"
    assert not os.path.exists(fname2)
    event = Event.from_data(data, fname=fname2)
    assert os.path.exists(fname2)
    event2 = Event.from_file(fname2)
    assert event.to_dict() == event2.to_dict()

    data = {
        (1, 1): [[0.35, 0.55]],  # since (1,1) exists, it won't be overwritten
        (2, 3): [[0.2, 0.33], [0.4, 0.516]],
    }
    # event will contain the data that was passed in, along with the data in the file named fname2
    event = Event.from_data(data, fname=fname2)
    # data from the original file is preserved is the key (1,1) already exists
    assert np.allclose(event.to_dict()[1, 1], [[0.3, 0.4]])
    event2 = Event.from_file(fname2)
    assert event.to_dict() == event2.to_dict()

    # with empty data, size must be specified
    data = {
        (1, 1): [],
        (1, 2): [],
        (2, 1): [],
        (2, 2): [],
    }
    with pytest.raises(AssertionError):
        event = Event.from_data(data, name="test_event")
    event = Event.from_data(data, name="test_event", size=2)


def test_event_to_portions():
    data = {
        (1, 1): [[0.1, 0.2], [0.3, 0.4]],
        (1, 2): [[0.5, 0.6]],
        (2, 1): [[0.7, 0.8], [0.9, 1.0], [0.85, 0.87]],
        (2, 2): [],
    }
    event = Event.from_data(data, name="test_event", pick_action="append")
    assert list(event.to_portions()[2, 2]) == []
    assert np.allclose(event.to_portions()[(2, 1)].duration, 0.22)

    data = {
        (1, 1): [[0.1], [0.3]],
        (1, 2): [[0.5]],
        (2, 1): [0.7, 0.9, 0.85],
        (2, 2): [],
    }
    event = Event.from_data(data, name="test_event", pick_action="append")
    assert event._data[(2, 1)].default == [[0.7], [0.9], [0.85]]
    with pytest.raises(AssertionError):
        event.to_portions()  # only works for 2-events


def test_events_initialization(matplotlib_figure):
    figure, ax = matplotlib_figure
    events = Events(figure)
    assert events.parent == figure


def test_events_add(matplotlib_figure):
    figure, ax = matplotlib_figure
    events = Events(figure)
    data_id_func = lambda: "test_id"
    event = events.add(
        name="test_event",
        size=2,
        fname="test.json",
        data_id_func=data_id_func,
        color="blue",
    )
    assert event.name == "test_event"
    assert event.size == 2


def test_events_add_from_file(matplotlib_figure, event_file_path):
    figure, ax = matplotlib_figure
    events = Events(figure)
    event = events.add_from_file(fname=event_file_path, data_id_func=lambda: "test_id")
    # same event as test_event_save
    assert np.allclose(
        event.to_dict()["test_id"], [[0.1, 0.17], [0.2, 0.25], [0.4, 0.6], [0.8, 0.85]]
    )
    events.setup_display()
    events.update_display()

    events = Events(figure)
    event = events.add_from_file(
        fname=event_file_path, data_id_func=lambda: "test_id", display_type="fill"
    )
    events.setup_display()
    events.update_display()

    events = Events(figure)
    event = events.add_from_file(
        fname=event_file_path, data_id_func=lambda: "test_id", ax_list=[ax]
    )
    events.setup_display()
    events.update_display()

    events = Events(figure)
    event = events.add_from_file(
        fname=event_file_path,
        data_id_func=lambda: "test_id",
        display_type="fill",
        ax_list=[ax],
    )
    events.setup_display()
    events.update_display()


def test_find_nearest_idx_val():
    # Test with a simple array
    array = [1.0, 2.0, 3.0, 4.0, 5.0]
    value = 3.3
    idx, val = _find_nearest_idx_val(array, value)
    assert idx == 2
    assert val == 3.0

    # Test with an array containing negative values
    array = [-5.0, -3.0, -1.0, 1.0, 3.0, 5.0]
    value = -2.5
    idx, val = _find_nearest_idx_val(array, value)
    assert idx == 1
    assert val == -3.0

    # Test with an array containing duplicate values
    array = [1.0, 2.0, 2.0, 3.0, 4.0]
    assert np.allclose(_find_nearest_idx_val(array, 2.4), (1, 2.0))
    assert np.allclose(_find_nearest_idx_val(array, 2.6), (3, 3.0))

    # Test with an empty array
    array = []
    value = 1.0
    with pytest.raises(ValueError):
        _find_nearest_idx_val(array, value)

    # Test with a single-element array
    array = [1.0]
    value = 0.5
    idx, val = _find_nearest_idx_val(array, value)
    assert idx == 0
    assert val == 1.0



"""Utility functions and classes for video processing and plotting.

This module provides various utility functions and classes to assist with
video processing, plotting, and other miscellaneous tasks.
"""

from __future__ import annotations

import errno
import os
import sys
from pathlib import Path
from typing import List, Tuple, Union

import cv2 as cv
import numpy as np
from decord import VideoReader, cpu
from matplotlib import axes as maxes
from matplotlib import pyplot as plt


def ticks_from_times(
    times: List[float], tick_lim: Tuple[float, float]
) -> Tuple[List[float], List[float]]:
    """Generate x, y arrays to supply to plt.plot function to plot a set of x-values (times) as ticks.

    Args:
        times (List[float]): List of time values.
        tick_lim (Tuple[float, float]): Limits for the y-axis ticks.

    Returns:
        Tuple[List[float], List[float]]: x and y arrays for plotting ticks.
    """

    def nan_pad_x(inp: List[float]) -> List[float]:
        return [item for x in inp for item in (x, x, np.nan)]

    def nan_pad_y(ylim: Tuple[float, float], n: int) -> List[float]:
        return [item for y1, y2 in [ylim] * n for item in (y1, y2, np.nan)]

    return nan_pad_x(times), nan_pad_y(tick_lim, len(times))


class _List(list):
    """Extended list with additional methods to find next and previous elements."""

    def next(self, val: float) -> float:
        """Next element in the list closest to val.

        Args:
            val (float): Value to find the next element for.

        Returns:
            float: Next element in the list closest to val.
        """
        return min([x for x in self if x > val], default=max(self))

    def previous(self, val: float) -> float:
        """Previous element in the list closest to val.

        Args:
            val (float): Value to find the previous element for.

        Returns:
            float: Previous element in the list closest to val.
        """
        return max([x for x in self if x < val], default=min(self))


def _parse_fax(
    fax: Union[None, plt.Figure, maxes.Axes],
    ax_pos: Tuple[float, float, float, float] = (0.01, 0.01, 0.98, 0.98),
) -> Tuple[plt.Figure, maxes.Axes]:
    """Helper function to parse figure and axes.

    Args:
        fax (Union[None, plt.Figure, maxes.Axes]): Figure or axes handle.
        ax_pos (Tuple[float, float, float, float], optional): Position of the axes. Defaults to (0.01, 0.01, 0.98, 0.98).

    Returns:
        Tuple[plt.Figure, maxes.Axes]: Parsed figure and axes.
    """
    assert isinstance(fax, (type(None), plt.Figure, maxes.Axes))
    if fax is None:
        f = plt.figure()
        ax = f.add_axes(ax_pos)
    elif isinstance(fax, plt.Figure):
        f = fax
        ax = f.add_axes(ax_pos)
    else:
        f = fax.figure
        ax = fax
    return f, ax


def _parse_pos(
    pos: Union[str, Tuple[float, float, str, str]]
) -> Tuple[float, float, str, str]:
    """Helper function to parse position strings.

    Args:
        pos (Union[str, Tuple[float, float, str, str]]): Position string or tuple.

    Returns:
        Tuple[float, float, str, str]: Parsed position.
    """
    if isinstance(pos, str):
        updown, leftright = pos.replace("middle", "center").split(" ")
        assert updown in ("top", "center", "bottom")
        assert leftright in ("left", "center", "right")
        y = {"top": 1, "center": 0.5, "bottom": 0}[updown]
        x = {"left": 0, "center": 0.5, "right": 1}[leftright]
        pos = (x, y, updown, leftright)
    assert len(pos) == 4
    return pos


class TextView:
    """Show text array line by line."""

    def __init__(
        self,
        text: Union[List[str], dict],
        fax: Union[None, plt.Figure, maxes.Axes] = None,
        pos: Union[str, Tuple[float, float, str, str]] = "bottom left",
    ):
        """
        Args:
            text (Union[List[str], dict]): Array of strings or dictionary to display.
            fax (Union[None, plt.Figure, maxes.Axes], optional): Figure or axes handle. Defaults to None.
            pos (Union[str, Tuple[float, float, str, str]], optional): Position of the text. Defaults to "bottom left".
        """

        def rescale(xy: float, margin: float = 0.03) -> float:
            return (1 - 2 * margin) * xy + margin

        self.text = self.parse_text(text)
        self._text = None  # matplotlib text object
        self._pos = _parse_pos(pos)
        self.figure, self._ax = _parse_fax(
            fax, ax_pos=(rescale(self._pos[0]), rescale(self._pos[1]), 0.02, 0.02)
        )
        self.setup()
        self.update()

    def parse_text(self, text: Union[List[str], dict]) -> List[str]:
        """Parse text input into a list of strings.

        Args:
            text (Union[List[str], dict]): Text input.

        Returns:
            List[str]: Parsed text.
        """
        if isinstance(text, dict):
            text = [f"{key} - {val}" for key, val in text.items()]
        return text

    def setup(self) -> None:
        """Setup for showing the text."""
        self._ax.axis("off")
        plt.show(block=False)

    def update(self, new_text: Union[List[str], dict] = None) -> None:
        """Update the text view with new text.

        Args:
            new_text (Union[List[str], dict], optional): New text to display. Defaults to None.
        """
        if new_text is not None:
            self.text = self.parse_text(new_text)
        if self._text is not None:
            self._text.remove()
        x, y, va, ha = self._pos
        self._text = self._ax.text(
            x, y, "\n".join(self.text), va=va, ha=ha, family="monospace"
        )
        plt.draw()


def get_palette(
    palette_name: str = "Set2", n_colors: int = 10
) -> List[Tuple[float, float, float]]:
    """Get a color palette, with fallback if seaborn is not available.

    Args:
        palette_name (str, optional): Name of the palette. Defaults to "Set2".
        n_colors (int, optional): Number of colors. Defaults to 10.

    Returns:
        List[Tuple[float, float, float]]: List of RGB tuples.
    """
    try:
        import seaborn as sns

        return sns.color_palette(palette_name, n_colors=n_colors)
    except ModuleNotFoundError:
        palettes = {
            "Set2": [
                (0.40, 0.76, 0.65),
                (0.99, 0.55, 0.38),
                (0.55, 0.63, 0.79),
                (0.91, 0.54, 0.76),
                (0.65, 0.85, 0.33),
                (1.00, 0.85, 0.18),
                (0.90, 0.77, 0.58),
                (0.70, 0.70, 0.70),
                (0.40, 0.76, 0.65),
                (0.99, 0.55, 0.38),
            ]*np.ceil(n_colors/10)  # Repeat the palette to cover n_colors
        }
        return palettes[palette_name][:n_colors]


def is_video(vide_file: str):
    cap = cv.VideoCapture(vide_file)
    if cap.isOpened():
        ret, _ = cap.read()
        cap.release()
        return ret  # If we can read a frame, it's a video
    return False


class Video(VideoReader):
    """Extended VideoReader class with additional methods."""

    def __init__(
        self,
        uri: str,
        ctx=cpu(0),
        width: int = -1,
        height: int = -1,
        num_threads: int = 0,
        fault_tol: int = -1,
    ):
        """
        Args:
            uri (str): Path to the video file.
            ctx: Context for video decoding. Defaults to cpu(0).
            width (int, optional): Width of the video. Defaults to -1.
            height (int, optional): Height of the video. Defaults to -1.
            num_threads (int, optional): Number of threads for decoding. Defaults to 0.
            fault_tol (int, optional): Fault tolerance. Defaults to -1.
        """
        assert os.path.exists(uri) and is_video(uri)
        self.fname = uri
        self.name = Path(uri).stem
        super().__init__(uri, ctx, width, height, num_threads, fault_tol)

    def gray(self, frame_num: int) -> np.ndarray:
        """Convert a frame to grayscale.

        Args:
            frame_num (int): Frame number to convert.

        Returns:
            np.ndarray: Grayscale frame.
        """
        return cv.cvtColor(self[frame_num].asnumpy(), cv.COLOR_BGR2GRAY)


def removeprefix(s: str, prefix: str) -> str:
    """Remove the specified prefix from the string, if present.

    Args:
        s (str): The original string.
        prefix (str): The prefix to remove.

    Returns:
        str: The string with the prefix removed, if it starts with the prefix.
    """
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def removesuffix(s: str, suffix: str) -> str:
    """Remove the specified suffix from the string, if present.

    Args:
        s (str): The original string.
        suffix (str): The suffix to remove.

    Returns:
        str: The string with the suffix removed, if it ends with the suffix.
    """
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


### ------- FROM STACK OVERFLOW
# Sadly, Python fails to provide the following magic number for us.
ERROR_INVALID_NAME = 123
"""
Windows-specific error code indicating an invalid pathname.

See Also
----------
https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
    Official listing of all such codes.
"""


def is_pathname_valid(pathname: str) -> bool:
    """
    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = (
            os.environ.get("HOMEDRIVE", "C:")
            if sys.platform == "win32"
            else os.path.sep
        )
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError as exc:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?


def is_path_creatable(pathname: str) -> bool:
    """
    `True` if the current user has sufficient permissions to create the passed
    pathname; `False` otherwise.
    """
    # Parent directory of the passed path. If empty, we substitute the current
    # working directory (CWD) instead.
    dirname = os.path.dirname(pathname) or os.getcwd()
    return os.access(dirname, os.W_OK)


def is_path_exists_or_creatable(pathname: str) -> bool:
    """
    `True` if the passed pathname is a valid pathname for the current OS _and_
    either currently exists or is hypothetically creatable; `False` otherwise.

    This function is guaranteed to _never_ raise exceptions.
    """
    try:
        # To prevent "os" module calls from raising undesirable exceptions on
        # invalid pathnames, is_pathname_valid() is explicitly called first.
        return is_pathname_valid(pathname) and (
            os.path.exists(pathname) or is_path_creatable(pathname)
        )
    # Report failure on non-fatal filesystem complaints (e.g., connection
    # timeouts, permissions issues) implying this path to be inaccessible. All
    # other exceptions are unrelated fatal issues and should not be caught here.
    except OSError:
        return False


### ------- END FROM STACK OVERFLOW



import os
import pytest
from unittest.mock import patch, MagicMock

# Assuming conftest.py is in the same directory or accessible via pytest's discovery
from .conftest import (
    simulate_key_press_at_xy,
    press_browser_button,
)
from datanavigator import examples


@pytest.fixture
def example_video_path(setup_folders):
    clip_dir, _ = setup_folders
    return os.path.join(clip_dir, "example_video.mp4")


# --- Tests for get_example_video ---


@patch("urllib.request.urlopen")
def test_get_example_video_download(mock_urlopen, example_video_path):
    """Test video download when file doesn't exist."""
    # Mock the download response
    mock_response = MagicMock()
    mock_response.read.return_value = b"video_content"
    mock_urlopen.return_value.__enter__.return_value = mock_response

    # Ensure file does not exist initially
    assert not os.path.exists(example_video_path)

    # Call the function
    result_path = examples.get_example_video()

    # Assertions
    assert result_path == example_video_path
    assert os.path.exists(example_video_path)
    with open(example_video_path, "rb") as f:
        assert f.read() == b"video_content"
    mock_urlopen.assert_called_once()


@patch("urllib.request.urlopen")
def test_get_example_video_exists(mock_urlopen, example_video_path):
    """Test returning path when video file already exists."""
    # Create a dummy existing file
    with open(example_video_path, "w") as f:
        f.write("existing_content")

    # Call the function
    result_path = examples.get_example_video()

    # Assertions
    assert result_path == example_video_path
    mock_urlopen.assert_not_called()  # Should not attempt download


def test_get_example_video_custom_folder_exists(tmp_path):
    """Test using a valid custom destination folder."""
    custom_folder = tmp_path / "custom_clips"
    custom_folder.mkdir()
    expected_path = custom_folder / "example_video.mp4"

    with patch("urllib.request.urlopen") as mock_urlopen:
        mock_response = MagicMock()
        mock_response.read.return_value = b"custom_content"
        mock_urlopen.return_value.__enter__.return_value = mock_response

        result_path = examples.get_example_video(dest_folder=str(custom_folder))

        assert result_path == str(expected_path)
        assert os.path.exists(result_path)
        mock_urlopen.assert_called_once()


def test_get_example_video_custom_folder_not_exist(tmp_path):
    """Test error when custom destination folder doesn't exist."""
    custom_folder = tmp_path / "non_existent_clips"
    with pytest.raises(AssertionError, match="does not exist"):
        examples.get_example_video(dest_folder=str(custom_folder))


@patch("urllib.request.urlopen")
def test_get_example_video_custom_url(mock_urlopen, example_video_path):
    """Test using a custom source URL."""
    custom_url = "http://example.com/custom_video.mp4"
    mock_response = MagicMock()
    mock_response.read.return_value = b"custom_url_content"
    mock_urlopen.return_value.__enter__.return_value = mock_response

    result_path = examples.get_example_video(source_url=custom_url)

    assert result_path == example_video_path
    assert os.path.exists(example_video_path)
    mock_urlopen.assert_called_once()
    # Check if the request used the custom URL
    mock_request = patch("urllib.request.Request").start()
    mock_request.return_value = MagicMock()
    mock_request.return_value.call_args = ((custom_url,), {})
    args, kwargs = mock_request.return_value.call_args
    assert args[0] == custom_url
    patch.stopall()


# --- Tests for EventPickerDemo ---


@pytest.fixture
def event_picker():
    """Fixture to create an EventPickerDemo instance."""
    # Prevent plt.show() from blocking tests
    with patch("matplotlib.pyplot.show"):
        picker = examples.EventPickerDemo()
    # Need the figure/axes from the picker for event simulation
    picker.fax = (picker.figure, picker._ax)
    return picker


def test_event_picker_init(event_picker):
    """Test EventPickerDemo initialization."""
    assert len(event_picker.data) == 10
    assert "pick1" in event_picker.events.names
    assert "pick2" in event_picker.events.names
    assert "pick3" in event_picker.events.names
    assert len(event_picker._ax.get_lines()) > 0  # Check if data is plotted


def test_event_picker_add_event(event_picker, setup_folders):
    """Test adding events via simulated key presses."""
    _, cache_dir = setup_folders
    picker = event_picker
    fax = picker.fax
    canvas = picker.figure.canvas

    # Add pick1 event
    event1_key = simulate_key_press_at_xy(fax, key="1", xdata=0.5)
    canvas.callbacks.process(event1_key.name, event1_key)
    current_id = picker.events["pick1"].data_id_func()
    assert current_id in picker.events["pick1"]._data
    assert len(picker.events["pick1"]._data[current_id].added) == 1
    assert picker.events["pick1"]._data[current_id].added[0] == [0.5]

    # Add pick2 event (requires two clicks)
    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.2))
    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.8))
    current_id = picker.events[
        "pick2"
    ].data_id_func()  # ID might change if signal changes
    assert current_id in picker.events["pick2"]._data
    assert len(picker.events["pick2"]._data[current_id].added) == 1
    assert picker.events["pick2"]._data[current_id].added[0] == [0.2, 0.8]

    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.3))
    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.25))
    assert (
        len(picker.events["pick2"]._data[current_id].added) == 1
    )  # Still 1 because secind click is before first click

    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.1))
    picker(simulate_key_press_at_xy(fax, key="2", xdata=0.15))
    assert (
        len(picker.events["pick2"]._data[current_id].added) == 2
    )  # Still 1 because secind click is before first click

    # Add pick3 event (overwrite) - requires three clicks
    event3_key_1 = simulate_key_press_at_xy(fax, key="3", xdata=0.1)
    event3_key_2 = simulate_key_press_at_xy(fax, key="3", xdata=0.5)
    event3_key_3 = simulate_key_press_at_xy(fax, key="3", xdata=0.9)
    canvas.callbacks.process(event3_key_1.name, event3_key_1)
    canvas.callbacks.process(event3_key_2.name, event3_key_2)
    canvas.callbacks.process(event3_key_3.name, event3_key_3)
    current_id = picker.events["pick3"].data_id_func()
    assert current_id in picker.events["pick3"]._data
    assert len(picker.events["pick3"]._data[current_id].added) == 1
    assert picker.events["pick3"]._data[current_id].added[0] == [0.1, 0.5, 0.9]

    # Add another pick3 event to test overwrite
    event3_key_4 = simulate_key_press_at_xy(fax, key="3", xdata=0.15)
    event3_key_5 = simulate_key_press_at_xy(fax, key="3", xdata=0.55)
    event3_key_6 = simulate_key_press_at_xy(fax, key="3", xdata=0.95)
    canvas.callbacks.process(event3_key_4.name, event3_key_4)
    canvas.callbacks.process(event3_key_5.name, event3_key_5)
    canvas.callbacks.process(event3_key_6.name, event3_key_6)
    assert (
        len(picker.events["pick3"]._data[current_id].added) == 1
    )  # Still 1 due to overwrite
    assert picker.events["pick3"]._data[current_id].added[0] == [0.15, 0.55, 0.95]


def test_event_picker_remove_event(event_picker):
    """Test removing an event via simulated key press."""
    picker = event_picker
    fax = picker.fax
    canvas = picker.figure.canvas

    # Add a pick1 event first
    add_event = simulate_key_press_at_xy(fax, key="1", xdata=0.6)
    canvas.callbacks.process(add_event.name, add_event)
    current_id = picker.events["pick1"].data_id_func()
    assert len(picker.events["pick1"]._data[current_id].added) == 1

    # Simulate remove key press near the added event
    remove_event = simulate_key_press_at_xy(
        fax, key="alt+1", xdata=0.605
    )  # Close to 0.6
    canvas.callbacks.process(remove_event.name, remove_event)

    # Check if the event was removed from 'added'
    assert len(picker.events["pick1"]._data[current_id].added) == 0


def test_event_picker_save_event(event_picker, setup_folders):
    """Test saving events via simulated key press."""
    _, cache_dir = setup_folders
    picker = event_picker
    fax = picker.fax
    canvas = picker.figure.canvas
    pick1_file = os.path.join(cache_dir, "_pick1.json")

    # Add a pick1 event
    add_event = simulate_key_press_at_xy(fax, key="1", xdata=0.7)
    canvas.callbacks.process(add_event.name, add_event)

    # Ensure save file doesn't exist yet
    assert not os.path.exists(pick1_file)

    # Simulate save key press
    save_event = simulate_key_press_at_xy(fax, key="ctrl+1")
    canvas.callbacks.process(save_event.name, save_event)

    # Check if the file was created
    assert os.path.exists(pick1_file)
    # Optional: Add check for file content


# --- Tests for ButtonDemo ---


@pytest.fixture
def button_demo():
    """Fixture to create a ButtonDemo instance."""
    # Prevent plt.show() from blocking tests
    with patch("matplotlib.pyplot.show"):
        demo = examples.ButtonDemo()
    return demo


def test_button_demo_init(button_demo):
    """Test ButtonDemo initialization."""
    assert len(button_demo.buttons._list) == 2
    assert button_demo.buttons["test"].__class__.__name__ == "ToggleButton"
    assert button_demo.buttons["push button"].__class__.__name__ == "Button"


@patch("builtins.print")
def test_button_demo_push_button_click(mock_print, button_demo):
    """Test clicking the push button."""
    push_button = button_demo.buttons["push button"]
    press_browser_button(push_button)
    # Check if the callback (which prints) was called
    mock_print.assert_called_once()


# --- Tests for SelectorDemo ---


@pytest.fixture
def selector_demo(matplotlib_figure):
    """Fixture to create a SelectorDemo instance."""
    # Prevent plt.show() from blocking tests
    with patch("matplotlib.pyplot.show"):
        # Need a figure with axes for SelectorDemo
        fig, ax = matplotlib_figure
        # Mock the figure creation within SelectorDemo if needed, or pass one
        with patch("matplotlib.pyplot.subplots", return_value=(fig, ax)):
            demo = examples.SelectorDemo()
    return demo


def test_selector_demo_init(selector_demo):
    """Test SelectorDemo initialization."""
    assert len(selector_demo.buttons._list) == 2
    assert selector_demo.lasso is not None  # Should be started by default
    assert len(selector_demo.ax.get_lines()) == 4  # data + selected points placeholder


def test_selector_demo_start_stop(selector_demo):
    """Test starting and stopping the lasso selector via buttons."""
    # Stop
    stop_button = selector_demo.buttons["Stop selection"]
    with patch.object(selector_demo.lasso, "disconnect_events") as mock_disconnect:
        press_browser_button(stop_button)
        mock_disconnect.assert_called_once()

    # Start
    start_button = selector_demo.buttons["Start selection"]
    with patch(
        "matplotlib.widgets.LassoSelector.__init__", return_value=None
    ) as mock_lasso_init:
        # Need to mock the LassoSelector init as it requires interaction
        with patch.object(
            examples.LassoSelectorWidget, "__init__", return_value=None
        ) as mock_lasso_widget_init:
            press_browser_button(start_button)
            # Check if LassoSelector was potentially re-initialized (hard to assert precisely without more mocking)
            assert mock_lasso_widget_init.called



import pytest
import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from unittest.mock import MagicMock
from datanavigator.core import GenericBrowser

from tests.conftest import simulate_key_press


def test_browser_initialization():
    GenericBrowser()
    GenericBrowser(figure_handle=plt.figure())


class TestGenericBrowser:
    @pytest.fixture
    def browser(self, matplotlib_figure):
        figure, ax = matplotlib_figure
        b = GenericBrowser(figure_handle=figure)
        b.events.add(
            name="test_event",
            size=2,
            fname="test.json",
            data_id_func=lambda: "test_id",
            color="blue",
        )
        b.statevariables.add(name="test_state", states=["state1", "state2"])
        b.statevariables.show()
        b.buttons.add(text="test_button", type_="Push", action_func=lambda: None)
        return b

    def test_initialization(self, browser):
        assert isinstance(browser.figure, plt.Figure)
        assert browser._current_idx == 0
        assert browser._keypressdict == {}
        assert browser._bindings_removed == {}

    def test_add_key_binding(self, browser):
        def dummy_function():
            pass

        browser.add_key_binding("alt+1", dummy_function, "test description")
        assert "alt+1" in browser._keypressdict
        assert browser._keypressdict["alt+1"] == (dummy_function, "test description")

    def test_set_default_keybindings(self, browser):
        browser.set_default_keybindings()
        assert "left" in browser._keypressdict
        assert "right" in browser._keypressdict
        assert "up" in browser._keypressdict
        assert "down" in browser._keypressdict

    def test_call_key_press_event(self, browser):
        event = simulate_key_press(browser.figure, key="left")
        browser.set_default_keybindings()
        browser._keypressdict["left"] = (MagicMock(), "decrement")
        browser._keypressdict["left"][0].assert_not_called()
        browser(event)
        browser._keypressdict["left"][0].assert_called_once()

    def test_call_close_event(self, browser):
        event = MagicMock()
        event.name = "close_event"
        browser.cleanup = MagicMock()
        browser.cleanup.assert_not_called()
        browser(event)
        browser.cleanup.assert_called_once()

    def test_update(self, browser):
        browser.update_assets = MagicMock()
        browser.update()
        browser.update_assets.assert_called_once()

    def test_update_without_clear(self, browser):
        browser.update_assets = MagicMock()
        browser.update_without_clear()
        browser.update_assets.assert_called_once()

    def test_reset_axes(self, browser):
        ax = browser.figure.get_axes()[0]
        ax.plot([0, 1], [0, 1])
        browser.reset_axes()
        assert np.allclose(ax.get_xlim(), (-0.05, 1.05))
        assert np.allclose(ax.get_ylim(), (-0.05, 1.05))

    def test_increment(self, browser):
        browser.data = [1, 2, 3, 4, 5]
        browser.update = MagicMock()
        browser.increment()
        assert browser._current_idx == 1
        browser.update.assert_called_once()

    def test_decrement(self, browser):
        browser.data = [1, 2, 3, 4, 5]
        browser.update = MagicMock()
        browser._current_idx = 5
        browser.decrement()
        assert browser._current_idx == 4
        browser.update.assert_called_once()

    def test_go_to_start(self, browser):
        browser.update = MagicMock()
        browser._current_idx = 5
        browser.go_to_start()
        assert browser._current_idx == 0
        browser.update.assert_called_once()

    def test_go_to_end(self, browser):
        browser.data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        browser.update = MagicMock()
        browser.go_to_end()
        assert browser._current_idx == 9
        browser.update.assert_called_once()

    def test_increment_frac(self, browser):
        browser.data = list(range(1000))
        browser.go_to_start()
        browser.update = MagicMock()
        browser.increment_frac()
        assert browser._current_idx == 50  # 20 steps to browse 1000 items
        browser.update.assert_called_once()
        browser._current_idx = 980
        browser.increment_frac()
        assert browser._current_idx == 999

    def test_decrement_frac(self, browser):
        browser.data = list(range(1000))
        browser.update = MagicMock()
        browser._current_idx = 10
        browser.decrement_frac()
        assert browser._current_idx == 0
        browser.update.assert_called_once()
        browser.go_to_end()
        browser.decrement_frac()
        assert browser._current_idx == 949

    def test_pan(self, browser):
        ax = browser.figure.get_axes()[0]
        ax.plot([0, 1], [0, 1])
        ax.set_xlim(0, 1)
        browser.pan(direction="left")
        assert np.allclose(ax.get_xlim(), (-0.2, 0.8))
        ax.set_ylim(-1, 1)
        browser.pan(direction="down")
        assert np.allclose(ax.get_ylim(), (-0.6, 1.4))

    def test_cleanup(self, browser):
        browser.figure.canvas.mpl_disconnect = MagicMock()
        browser.mpl_restore_bindings = MagicMock()
        browser.cleanup()
        browser.figure.canvas.mpl_disconnect.assert_called()
        browser.mpl_restore_bindings.assert_called_once()

    def test_mpl_restore_bindings(self, browser):
        browser._bindings_removed = {"keymap.back": "left"}
        browser.mpl_restore_bindings()
        assert "left" in mpl.rcParams["keymap.back"]

    def test_has(self, browser):
        browser.buttons = [1]
        assert browser.has("buttons")
        browser.buttons = []
        assert not browser.has("buttons")

    def test_filter_sibling_axes(self, browser):
        ax1 = browser.figure.add_subplot(211)
        ax2 = browser.figure.add_subplot(212, sharex=ax1)
        result = browser._filter_sibling_axes([ax1, ax2], share="x")
        assert result == [ax1]

    def test_memoryslots(self, browser):
        event = simulate_key_press(browser.figure, key="1")
        browser.data = [1, 2, 3, 4, 5]
        browser._current_idx = 2
        browser(event)
        assert browser.memoryslots._list["1"] == 2

    def test_show_key_bindings(self, browser):
        browser.show_key_bindings("new")



from __future__ import annotations

import functools
import json
import os
from pathlib import Path
from typing import Callable, Mapping, Union, List, Tuple, Dict, Optional, Any

import numpy as np
import pandas as pd
import pysampled
from matplotlib import pyplot as plt
from matplotlib.animation import FFMpegWriter

from datanavigator import utils
from datanavigator.assets import AssetContainer
from datanavigator.videos import VideoBrowser
from datanavigator.opticalflow import lucas_kanade, lucas_kanade_rstc


class VideoPointAnnotator(VideoBrowser):
    """
    Add point annotations to videos.

    Use arrow keys to navigate frames.
    Select a 'label category' from 0 to 9 by pressing the corresponding number key.
    Point your mouse at a desired location in the video and press the forward slash / button to add a point annotation.
    When you're done, press 's' to save your work, which will create a '<video name>_annotations.json' file in the same folder as the video file.
    These annotations will be automagically loaded when you try to annotate this file again.

    If you're doing one label at a time, then pick the frames for the first label arbitrarily.
    For the second label onwards,

    Args:
        vid_name (Path): Path to the video.
        annotation_names (Union[List[str], Mapping[str, Path]], optional):
            List[str] - Name(s) of the annotation layer(s). The file path(s) are deduced from the name(s).
            Mapping[str, Path] - A dictionary mapping of annotation layer names to the annotation file paths.
            Defaults to '' with one layer of annotations.
        titlefunc (Callable, optional): A function used to set the title of the image axis.
            Defaults to a title function specified in :py:class:`VideoBrowser`.
    """

    def __init__(
        self,
        vid_name: Path,
        annotation_names: Union[
            List[str], Mapping[str, Path], List[VideoAnnotation]
        ] = "",
        n_labels: int = 10,
        titlefunc: Callable = None,
        image_process_func: Callable = lambda im: im,
        height_ratios: Tuple = (10, 1, 1),  # depends on your screen size
    ) -> None:
        figure_handle = plt.figure(constrained_layout=True, figsize=(12, 8))
        gs = figure_handle.add_gridspec(3, 2, width_ratios=[1, 4], height_ratios=list(height_ratios))
        self._ax_image = figure_handle.add_subplot(gs[0, 1])
        self._ax_trace_x = figure_handle.add_subplot(gs[1, :])
        self._ax_trace_y = figure_handle.add_subplot(gs[2, :])
        self._ax_trace_x.sharex(self._ax_trace_y)
        (self._frame_marker_x,) = self._ax_trace_x.plot(
            [], [], color="black", linewidth=1
        )
        (self._frame_marker_y,) = self._ax_trace_y.plot(
            [], [], color="black", linewidth=1
        )
        super().__init__(vid_name, titlefunc, self._ax_image, image_process_func)
        self.memoryslots.hide()
        self.memoryslots.disable()

        # annotation layers
        self.annotations = VideoAnnotations(parent=self)
        self.add_annotation_layers(annotation_names, n_labels)
        if "buffer" in self.annotations.names:
            self.annotations["buffer"].plot_type = "line"

        # frames of interest
        self.frames_of_interest = []
        (self._plot_frames_of_interest_x,) = self._ax_trace_x.plot(
            [], [], color="gray", linewidth=1, alpha=0.5
        )
        (self._plot_frames_of_interest_y,) = self._ax_trace_y.plot(
            [], [], color="gray", linewidth=1, alpha=0.5
        )

        # State variables
        self.statevariables.add("annotation_layer", self.annotations.names)
        self.statevariables.add("annotation_overlay", [None] + self.annotations.names)
        self.statevariables.add("annotation_label", self.ann.labels)
        self.statevariables.add("label_range", [f"{x*10}-{x*10+9}" for x in range(100)]) # up to 1000 labels
        first_label = self.ann.labels[0]
        self.statevariables["label_range"].set_state(int(first_label) // 10)
        self.update_annotation_label_states()
        self.statevariables.add("number_keys", ["select", "place"])
        self._ax_statevar = figure_handle.add_subplot(gs[0, 0])
        self.statevariables.show(pos="middle left", fax=self._ax_statevar)

        self.add_events()

        self.set_key_bindings()

        # set mouse click behavior
        self.cid.append(
            self.figure.canvas.mpl_connect("pick_event", self.select_label_with_mouse)
        )
        self.cid.append(
            self.figure.canvas.mpl_connect(
                "button_press_event", self.place_label_with_mouse
            )
        )
        self.cid.append(
            self.figure.canvas.mpl_connect("button_press_event", self.go_to_frame)
        )

        if self.__class__.__name__ == "VideoPointAnnotator":
            plt.show(block=False)
            self.update()
            plt.setp(self._ax_trace_x.get_xticklabels(), visible=False)
            plt.draw()

    @classmethod
    def from_annotations(
        cls, annotations: List[VideoAnnotation], *args, **kwargs
    ) -> VideoPointAnnotator:
        if isinstance(annotations, VideoAnnotation):
            annotations = [annotations]
        video_names = {a.video.fname for a in annotations}
        assert len(video_names) == 1  # same video across all annotations
        return cls(video_names.pop(), annotations, *args, **kwargs)

    def add_annotation_layers(
        self, 
        annotation_names: Union[List[str], Dict[str, Path], List[VideoAnnotation]],
        n_labels: int = 10
    ) -> None:
        """Load data from annotation files if they exist, otherwise initialize annotation layers."""
        if isinstance(annotation_names, (str, VideoAnnotation)):
            annotation_names = [annotation_names]

        if isinstance(annotation_names, list) and all(
            [isinstance(a, VideoAnnotation) for a in annotation_names]
        ):
            annotation_names = {
                a.name: a.fname for a in annotation_names
            }  # re-add because of plotting

        if "buffer" not in annotation_names and "buffer" not in self.annotations.names:
            if isinstance(annotation_names, list):
                annotation_names.append("buffer")
            else:
                annotation_names["buffer"] = self._get_fname_annotations("buffer")

        if isinstance(annotation_names, dict):
            ann_name_fname = annotation_names
        else:
            ann_name_fname = {
                name: self._get_fname_annotations(name) for name in annotation_names
            }

        for name, fname in ann_name_fname.items():
            self.annotations.add(
                name=name,
                fname=fname,
                vname=self.fname,
                ax_list_scatter=[self._ax_image],
                ax_list_trace_x=[self._ax_trace_x],
                ax_list_trace_y=[self._ax_trace_y],
                palette_name="Set2",
                n_labels=n_labels,
            )

        # same set of non-empty labels in all the loaded annotations
        all_labels = []
        for ann in self.annotations._list:
            all_labels += [
                label for label, label_data in ann.data.items() if label_data
            ]
        all_labels = sorted(list(set(all_labels)))
        if not all_labels:
            # when starting without any annotations, initialize a full set of empty annotations
            all_labels = [str(x) for x in range(n_labels)]
        for ann in self.annotations._list:
            for label in all_labels:
                if label not in ann.labels:
                    ann.add_label(label)
            for label in ann.labels:
                if label not in all_labels:
                    assert not ann.data[label]
                    del ann.data[label]
            ann.sort_labels()
            ann.re_setup_display()

        if "annotation_layer" in self.statevariables.names:
            self.statevariables["annotation_layer"].states = self.annotations.names
        if "annotation_overlay" in self.statevariables.names:
            self.statevariables["annotation_overlay"].states = [
                None
            ] + self.annotations.names
    

    def set_key_bindings(self) -> None:
        """Set the keyboard actions."""
        self.add_key_binding("s", self.save, "Save current annotation layer")
        self.add_key_binding("t", self.add_annotation)
        self.add_key_binding("y", self.remove_annotation)

        self.add_key_binding("f", self.increment_if_unannotated)
        self.add_key_binding("g", self.increment)
        self.add_key_binding("d", self.decrement_if_unannotated)

        self.add_key_binding("-", self.previous_annotation_layer)
        self.add_key_binding("=", self.next_annotation_layer)
        self.add_key_binding("[", self.previous_annotation_overlay)
        self.add_key_binding("]", self.next_annotation_overlay)
        self.add_key_binding(";", self.previous_annotation_label)
        self.add_key_binding("'", self.next_annotation_label)
        self.add_key_binding(",", self.previous_frame_with_any_label)
        self.add_key_binding(".", self.next_frame_with_any_label)
        self.add_key_binding("alt+,", self.previous_frame_of_interest)
        self.add_key_binding("alt+.", self.next_frame_of_interest)

        self.add_key_binding(
            "j",
            (lambda s: s.pan(direction="left")).__get__(self),
            description="pan left",
        )
        self.add_key_binding(
            "k",
            (lambda s: s.pan(direction="right")).__get__(self),
            description="pan right",
        )

        self.add_key_binding("`", self.cycle_number_keys_behavior)

        self.add_key_binding("n", self.next_frame_with_current_label)
        self.add_key_binding("p", self.previous_frame_with_current_label)
        self.add_key_binding("b", self.previous_frame_with_current_label)

        self.add_key_binding("m", self.toggle_frame_of_interest)
        self.add_key_binding("c", self.copy_current_annotation_from_overlay)
        # self.add_key_binding('ctrl+alt+c', self.copy_annotations_from_overlay)
        self.add_key_binding("alt+c", self.copy_frames_of_interest_from_overlay)
        self.add_key_binding("ctrl+alt+c", self.copy_frames_in_interval_from_overlay)

        self.add_key_binding(
            "v",
            (lambda s: s.check_labels_with_lk(mode="minimal")).__get__(self),
            "Check labels with LK - minimal mode",
        )
        self.add_key_binding(
            "alt+v",
            (lambda s: s.check_labels_with_lk(mode="current")).__get__(self),
            "Check labels with LK - current label",
        )
        self.add_key_binding(
            "ctrl+alt+v",
            (lambda s: s.check_labels_with_lk(mode="all")).__get__(self),
            "Check labels with LK - all labels",
        )

        self.add_key_binding(
            "a", self.interpolate_with_lk, "Interpolate current point with LK"
        )
        self.add_key_binding(
            "ctrl+a",
            (lambda s: s.interpolate_with_lk(all_labels=True)).__get__(self),
            "Interpolate all points with LK",
        )

        self.add_key_binding(
            "ctrl+d",
            (lambda s: s.interpolate_with_lk_norstc(all_labels=True)).__get__(self),
            "Interpolate all points with LK",
        )

        self.add_key_binding(
            "alt+a", self.remove_labels_in_interval, "Clear current label in interval"
        )
        self.add_key_binding(
            "ctrl+alt+a",
            (lambda s: s.remove_labels_in_interval(all_labels=True)).__get__(self),
            "Clear all labels in interval",
        )

        self.add_key_binding(
            "alt+b",
            (lambda s: s.predict_points_with_lucas_kanade(labels="current")).__get__(
                self
            ),
            "Predict current point with lucas-kanade",
        )
        self.add_key_binding(
            "ctrl+b",
            (lambda s: s.predict_points_with_lucas_kanade(labels="all")).__get__(self),
            "Predict all points with lucas-kanade",
        )

        self.add_key_binding("alt+q", self.keep_overlapping_continuous_frames)

        self.add_key_binding("w", self.increment_label_range)
        self.add_key_binding("q", self.decrement_label_range)

        self.remove_key_binding("e") # remove the "Extract clip feature from VideoBrowser

    def add_events(self) -> None:
        """Add an event to specify time intervals for interpolating with lucas-kanade."""
        event_name = "interp_with_lk"
        self.events.add(
            name=event_name,
            size=2,
            fname=os.path.join(
                Path(self.fname).parent,
                Path(self.fname).stem + f"_events_{event_name}.json",
            ),
            data_id_func=(lambda s: (s._current_layer, s._current_label)).__get__(self),
            data_func=round,
            color="gray",
            pick_action="overwrite",
            ax_list=[self._ax_trace_x],
            add_key="z",
            remove_key=None,
            save_key=None,
            display_type="fill",
            win_remove=(10, 10),
            show=True,
        )

    @property
    def ann(self) -> VideoAnnotation:
        """Return current annotation layer."""
        return self.annotations[self._current_layer]

    @property
    def _current_label(self) -> str:
        """Return current label."""
        return str(int(self.statevariables["annotation_label"].current_state) + int(self.statevariables["label_range"]._current_state_idx) * 10)

    @property
    def _current_layer(self) -> str:
        """Return current annotation layer"""
        return self.statevariables["annotation_layer"].current_state

    @property
    def _current_overlay(self) -> Union[str, None]:
        """Return current annotation overlay layer"""
        return self.statevariables["annotation_overlay"].current_state

    def _get_fname_annotations(
        self, annotation_name: str, suffix: str = ".json"
    ) -> str:
        """Construct the filename corresponding to an  annotation layer named annotation_name."""
        return os.path.join(
            Path(self.fname).parent,
            Path(self.fname).stem
            + "_annotations"
            + f'{"_" if annotation_name else ""}'
            + annotation_name
            + suffix,
        )

    def __call__(self, event: Any) -> None:
        """Callbacks for number keys."""
        super().__call__(event)
        # if a number key is pressed
        if event.name == "key_press_event" and str(event.key).isdigit() and int(event.key) in range(10):
            key_str = str(event.key)
            key_int = int(event.key)
            label = str(key_int + int(self.statevariables["label_range"]._current_state_idx) * 10)
            
            # if the label already exists
            if label in self.ann.labels:
                self.statevariables["annotation_label"].set_state(key_str)
                if self.statevariables["number_keys"].current_state == "place":
                    self.add_annotation(event)
            else:                
                # add a new label if the label doesn't exist
                for ann in self.annotations._list:  # add new label to all annotations
                    if label not in ann.labels:
                        ann.add_label(label)
                self.update_annotation_label_states()
                self.statevariables["annotation_label"].set_state(key_str)
                if self.statevariables["number_keys"].current_state == "place":
                    self.add_annotation(event)

            self.update()

    def update(self) -> None:
        """Update elements in the UI."""
        self.update_annotation_visibility(draw=False)
        # self.ann.update_display(self._current_idx, draw=False)
        # self.ann.show_one_trace(self._current_label, draw=False)
        self.statevariables.update_display(draw=False)
        self.update_frame_marker(draw=False)
        super().update()
        plt.draw()

    def update_annotation_visibility(self, draw: bool = False) -> None:
        """Update the visibility of all annotation layers, for example, when the layer is changed."""
        for ann in self.annotations:
            if ann.name == self._current_layer:
                ann.set_alpha(1, draw=False)
                ann.show(draw=False)
                ann.show_one_trace(self._current_label, draw=False)
                ann.update_display(self._current_idx, draw=draw)
            elif ann.name == self._current_overlay:
                if ann.name != self._current_layer:
                    ann.set_alpha(0.4, draw=False)
                    ann.show(draw=False)
                    ann.show_one_trace(self._current_label, draw=False)
                    ann.update_display(self._current_idx, draw=draw)
            else:
                ann.hide(draw=draw)

    def update_frame_marker(self, draw: bool = False) -> None:
        """Update the current frame location in the trace plots."""

        def nanlim(x, default):
            if np.all(np.isnan(x)):
                return default
            return [np.nanmin(x) * 0.9, np.nanmax(x) * 1.1]

        def nanlim_small(x, default, scale=0.6):
            if np.all(np.isnan(x)):
                nmin, nmax = default
            else:
                nmin = np.nanmin(x) * 0.9
                nmax = np.nanmax(x) * 1.1
            m = (nmin + nmax) / 2
            return [(nmin - m) * scale + m, (nmax - m) * scale + m]

        trace_data_x, trace_data_y = np.hstack(
            [ann.to_trace(self._current_label).T for ann in self.annotations._list]
        )

        default_x, default_y = self._ax_trace_x.get_ylim(), self._ax_trace_y.get_ylim()
        xl, yl = nanlim(trace_data_x, default_x), nanlim(trace_data_y, default_y)
        xls, yls = nanlim_small(trace_data_x, default_x), nanlim_small(trace_data_y, default_y)

        self._frame_marker_x.set_data([self._current_idx] * 2, xls)
        self._frame_marker_y.set_data([self._current_idx] * 2, yls)

        self._plot_frames_of_interest_x.set_data(
            *utils.ticks_from_times(self.frames_of_interest, xl)
        )
        self._plot_frames_of_interest_y.set_data(
            *utils.ticks_from_times(self.frames_of_interest, yl)
        )
        # self._ax_trace_x.set_xlim((0, n_frames))
        # if len(self.ann.data[self._current_label]) > 0:
        if not np.any(np.isnan(xl)):
            self._ax_trace_x.set_ylim(xl)
            self._ax_trace_y.set_ylim(yl)
        if draw:
            plt.draw()

    def copy_annotations_from_overlay(self) -> None:
        """Copy annotations from the overlay layer into the current layer in the current frame."""
        ann_overlay = self.annotations[self._current_overlay]
        frame_number = self._current_idx
        for label in self.ann.labels:
            if label in ann_overlay.labels:
                location = ann_overlay.data[label].get(frame_number, None)
                if location is not None:
                    self.ann.add(location, label, frame_number)
        self.update()

    def copy_current_annotation_from_overlay(self) -> None:
        """Copy annotations from the overlay layer into the current layer."""
        ann_overlay = self.annotations[self._current_overlay]
        frame_number = self._current_idx
        label = self._current_label
        if label in ann_overlay.labels:
            location = ann_overlay.data[label].get(frame_number, None)
            if location is not None:
                self.ann.add(location, label, frame_number)
        self.update()

    def copy_frames_of_interest_from_overlay(self) -> None:
        """copy annotations at frames of interest from buffer into the current layer.
        If there is no buffer, then copy from the overlay layer.
        """
        source_ann = self.annotations[self._current_overlay]

        for frame_number in self.frames_of_interest:
            for label in self.ann.labels:
                if label in source_ann.labels:
                    location = source_ann.data[label].get(frame_number, None)
                    if location is not None:
                        self.ann.add(location, label, frame_number)
        self.update()

    def copy_frames_in_interval_from_overlay(self) -> None:
        """For the current label only."""
        start_frame, end_frame = self.get_selected_interval()
        ann_overlay = self.annotations[self._current_overlay]
        label = self._current_label
        if label in ann_overlay.labels:
            for frame_number in range(start_frame, end_frame + 1):
                location = ann_overlay.data[label].get(frame_number, None)
                if location is not None:
                    self.ann.add(location, label, frame_number)
        self.update()

    def _add_annotation(
        self,
        location: List[float],
        frame_number: Optional[int] = None,
        label: Optional[str] = None,
    ) -> None:
        """Core function for adding annotations. Allows more control."""
        if frame_number is None:
            frame_number = self._current_idx
        if label is None:
            label = self._current_label
        self.ann.add(location, label, frame_number)

    def add_annotation(self, event: Any) -> None:
        """Add annotation at frame. If it exists, it'll get overwritten."""
        if event.inaxes == self._ax_image:
            self._add_annotation([float(event.xdata), float(event.ydata)])
        self.update()

    def remove_annotation(self, event: Optional[Any] = None) -> None:
        """remove annotation at the current frame if it exists"""
        self.ann.remove(self._current_label, self._current_idx)
        self.update()

    def _get_nearest_annotated_frame(self, label: Optional[str] = None) -> int:
        """Return the nearest frame (in either direction) number with the current label in the current annotation layer."""
        if label is None:
            label = self._current_label
        d = {abs(x - self._current_idx): x for x in self.ann.get_frames(label)}
        if not d:
            raise ValueError(
                f"No frames with label {label} in annotation layer {self._current_layer}."
            )
        return d[min(d)]

    def previous_frame_with_current_label(self, event: Optional[Any] = None) -> None:
        """Go to the previous frame with the current label in the current annotation layer."""
        self._current_idx = max(
            [
                x
                for x in self.ann.get_frames(self._current_label)
                if x < self._current_idx
            ],
            default=self._current_idx,
        )
        self.update()

    def next_frame_with_current_label(self, event: Optional[Any] = None) -> None:
        """Go to the next frame with the current label in the current annotation layer."""
        self._current_idx = min(
            [
                x
                for x in self.ann.get_frames(self._current_label)
                if x > self._current_idx
            ],
            default=self._current_idx,
        )
        self.update()

    def previous_frame_with_any_label(self, event: Optional[Any] = None) -> None:
        """Go to the previous frame with any label in the current annotation layer."""
        self._current_idx = max(
            [x for x in self.ann.frames if x < self._current_idx],
            default=self._current_idx,
        )
        self.update()

    def next_frame_with_any_label(self, event: Optional[Any] = None) -> None:
        """Go to the next frame with any label in the current annotation layer."""
        self._current_idx = min(
            [x for x in self.ann.frames if x > self._current_idx],
            default=self._current_idx,
        )
        self.update()

    def previous_frame_of_interest(self, event: Optional[Any] = None) -> None:
        """Go to the previous frame of interest."""
        self._current_idx = max(
            [x for x in self.frames_of_interest if x < self._current_idx],
            default=self._current_idx,
        )
        self.update()

    def next_frame_of_interest(self, event: Optional[Any] = None) -> None:
        self._current_idx = min(
            [x for x in self.frames_of_interest if x > self._current_idx],
            default=self._current_idx,
        )
        self.update()

    def previous_annotation_layer(self) -> None:
        """Go to the previous annotation layer."""
        self.statevariables["annotation_layer"].cycle_back()
        self.update()

    def next_annotation_layer(self) -> None:
        """Go to the next annotation layer"""
        self.statevariables["annotation_layer"].cycle()
        self.update()

    def previous_annotation_overlay(self) -> None:
        """Go to the previous annotation overlay layer."""
        self.statevariables["annotation_overlay"].cycle_back()
        self.update()

    def next_annotation_overlay(self) -> None:
        """Go to the next annotation overlay layer."""
        self.statevariables["annotation_overlay"].cycle()
        self.update()

    def previous_annotation_label(self) -> None:
        """Set current annotation label to the previous one."""
        self.statevariables["annotation_label"].cycle_back()
        self.update()

    def next_annotation_label(self) -> None:
        """Set current annotation label to the next one."""
        self.statevariables["annotation_label"].cycle()
        self.update()
    
    def update_annotation_label_states(self) -> None:
        """Update the states of the annotation label state variable."""
        # find labels in the current range
        label_states = [str(int(x) % 10) for x in self.ann.labels if int(x) in range(int(self.statevariables["label_range"]._current_state_idx) * 10, (int(self.statevariables["label_range"]._current_state_idx) + 1) * 10)]
        self.statevariables["annotation_label"].states = label_states

    def cycle_number_keys_behavior(self) -> None:
        """Number keys can be used to either select labels, or place a specific label.
        Toggle between these two behaviors.
        """
        self.statevariables["number_keys"].cycle()
        self.update()

    def increment_label_range(self) -> None:
        """Increment the label range by 10."""
        self.statevariables["label_range"].cycle()
        # if the current label is not present, add it to the list of labels in all the layers!
        current_label = self._current_label
        if self._current_label not in self.ann.labels:
            for ann in self.annotations._list:
                ann.add_label(self._current_label)
        self.update_annotation_label_states()
        self.statevariables["annotation_label"].set_state(str(int(current_label) % 10))
        self.update()
    
    def decrement_label_range(self) -> None:
        """Increment the label range by 10."""
        self.statevariables["label_range"].cycle_back()
        # if the current label is not present, add it to the list of labels
        current_label = self._current_label
        if self._current_label not in self.ann.labels:
            for ann in self.annotations._list:
                ann.add_label(self._current_label)
        self.update_annotation_label_states()
        self.statevariables["annotation_label"].set_state(str(int(current_label) % 10))
        self.update()

    def increment_if_unannotated(self, event: Optional[Any] = None) -> None:
        """Advance the frame if the current frame doesn't have any annotations.
        Useful to pause at annotated frames when adding a new label.
        """
        if self._current_idx not in self.ann.frames:
            self.increment()

    def decrement_if_unannotated(self, event: Optional[Any] = None) -> None:
        """Go to the previous frame if the current frame doesn't have any annotations.
        Useful to pause at annotated frames when adding a new label.
        """
        if self._current_idx not in self.ann.frames:
            self.decrement()

    def save(self) -> None:
        """Save current annotation layer json file."""
        self.ann.save()

    def select_label_with_mouse(self, event: Any) -> None:
        """Select a label by clicking on it with the left mousebutton."""
        if event.mouseevent.button.name == "LEFT":
            picked_label = self.ann.labels[int(event.ind[0])]
            self.statevariables["label_range"]._current_state_idx = int(picked_label) // 10
            self.update_annotation_label_states()
            self.statevariables["annotation_label"].set_state(str(int(picked_label) % 10))
            print(
                f'Picked label {picked_label} at frame {self._current_idx}'
            )
            self.update()

    def place_label_with_mouse(self, event: Any) -> None:
        """Place the selected label with the right mousebutton."""
        if event.inaxes == self._ax and event.button.name == "RIGHT":
            self.add_annotation(event)

    def go_to_frame(self, event: Any) -> None:
        if (
            event.inaxes in (self._ax_trace_x, self._ax_trace_y)
            and event.button.name == "RIGHT"
        ):
            self._current_idx = round(event.xdata)
            self.update()

    def toggle_frame_of_interest(self, event: Any) -> None:
        """Mark/unmark the current frame as a frame of interest"""
        if event.inaxes in (self._ax_trace_x, self._ax_trace_y):
            frame_number = self._current_idx
            if frame_number in self.frames_of_interest:
                self.frames_of_interest.remove(frame_number)
            else:
                self.frames_of_interest.append(frame_number)
            self.frames_of_interest.sort()
            self.update()

    def keep_overlapping_continuous_frames(self) -> None:
        self.ann.keep_overlapping_continuous_frames()
        self.update()

    def predict_points_with_lucas_kanade(
        self,
        labels: Union[str, List[str]] = "all",
        start_frame: Optional[int] = None,
        mode: str = "full",
    ) -> Any:
        """Compute the location of labels at the current frame using Lucas-Kanade algorithm."""
        if labels == "all":
            labels = self.ann.labels
        elif labels == "current":
            labels = [self._current_label]
        elif isinstance(labels, str):  # specify one label
            assert labels in self.ann.labels
            labels = [labels]
        else:  # specify a list of labels
            assert all([label in self.ann.labels for label in labels])

        if start_frame is None:
            start_frame = self._get_nearest_annotated_frame()

        assert (
            len(set([self._get_nearest_annotated_frame(label) for label in labels]))
            == 1
        ), "There must be a unique annotated frame across all labels."
        end_frame = self._current_idx  # always predict at the current location

        video = self.data
        init_loc = [self.ann.data[label][start_frame] for label in labels]
        tracked_loc = lucas_kanade(video, start_frame, end_frame, init_loc, mode=mode)
        end_loc_all = tracked_loc[-1]
        for label, end_loc in zip(labels, end_loc_all):
            if end_frame in self.ann.get_frames(label):
                print(f"Updating location for {label} at {end_frame}.")
                print(
                    f"To revert, use v._add_annotation({self.ann.data[label][end_frame]}, frame_number={end_frame}, label='{label}'); v.update()"
                )
            self._add_annotation(end_loc, label=label)
        self.update()
        return tracked_loc

    def get_selected_interval(self) -> Tuple[int, int]:
        start_frame, end_frame = (
            self.events["interp_with_lk"]
            ._data[(self._current_layer, self._current_label)]
            .get_times()[-1]
        )
        return start_frame, end_frame

    def remove_labels_in_interval(self, all_labels: bool = False) -> None:
        video = self.data
        if self._current_overlay is None:
            return

        if all_labels:
            label_list = self.annotations[self._current_overlay].labels
        else:
            label_list = [self._current_label]

        start_frame, end_frame = self.get_selected_interval()
        ann_overlay = self.annotations[self._current_overlay]
        for frame_count, frame_number in enumerate(range(start_frame, end_frame + 1)):
            for label in label_list:
                self.ann.remove(label, frame_number)
        self.update()

    def interpolate_with_lk(self, all_labels: bool = False) -> None:
        """Interpolate with lk-rstc between selected interval.
        Use data in the overlay layer as start and end points.
        Add interpolated points to the current layer.
        """
        video = self.data
        if self._current_overlay is None:
            return

        if all_labels:
            label_list = self.annotations[self._current_overlay].labels
        else:
            label_list = [self._current_label]

        start_frame, end_frame = self.get_selected_interval()
        ann_overlay = self.annotations[self._current_overlay]
        start_points = [ann_overlay.data[label][start_frame] for label in label_list]
        end_points = [ann_overlay.data[label][end_frame] for label in label_list]
        rstc_path = lucas_kanade_rstc(
            video, start_frame, end_frame, start_points, end_points
        )
        for frame_count, frame_number in enumerate(range(start_frame, end_frame + 1)):
            for label_count, label in enumerate(label_list):
                location = list(rstc_path[frame_count, label_count, :])
                self._add_annotation(location, frame_number, label)
        self.update()

    def interpolate_with_lk_norstc(self, all_labels: bool = False) -> None:
        """Infer the motion of the current label using lucas-kanade algorithm in the selected interval."""
        video = self.data
        if self._current_overlay is None:
            return

        if all_labels:
            label_list = self.annotations[self._current_overlay].labels
        else:
            label_list = [self._current_label]

        start_frame, end_frame = self.get_selected_interval()
        ann_overlay = self.annotations[self._current_overlay]
        start_points = [ann_overlay.data[label][start_frame] for label in label_list]
        # end_points = [ann_overlay.data[label][end_frame] for label in label_list]
        rstc_path = lucas_kanade(video, start_frame, end_frame, start_points)
        for frame_count, frame_number in enumerate(range(start_frame, end_frame + 1)):
            for label_count, label in enumerate(label_list):
                location = list(rstc_path[frame_count, label_count, :])
                self._add_annotation(location, frame_number, label)
        self.update()

    def check_labels_with_lk(self, mode: str = "minimal") -> None:
        """Interpolate between all labeled frames.
        This only makes sense for sparse-labeled annotations.
        Use this when doing first-time annotations (as opposed to refinement).

        I am testing if refining the start labels using this strategy,
        and augmenting the training data will improve deeplabcut tracking!

        Args:
            mode (str, optional):
                "all"     - LK-interpolation for all labels across all labeled frames
                "current" - LK-interpolation for current label across all labeled frames
                "minimal" - LK-interpolation for current label between labeled frames near the current frame (two previous to two next)
                Defaults to "minimal".
        """
        assert mode in ("current", "all", "minimal")

        source_ann = self.ann
        if "buffer" in self.annotations.names:
            target_ann = self.annotations["buffer"]

        video = source_ann.video
        if mode == "all":
            label_list = source_ann.labels
        else:
            label_list = [self._current_label]

        rstc_paths = {
            label: np.full((source_ann.n_frames, 2), np.nan) for label in label_list
        }
        for label in label_list:
            frames = utils._List(sorted(list(source_ann[label].keys())))
            if mode == "minimal":
                c = self._current_idx
                n1 = c if c in frames else frames.next(c)
                n2 = frames.next(n1)
                p1 = frames.previous(c)
                p2 = frames.previous(p1)
                frames = sorted(list(set([p2, p1, n1, n2])))
            for start_frame, end_frame in zip(frames, frames[1:]):
                start_points = [source_ann[label][start_frame]]
                end_points = [source_ann[label][end_frame]]
                rstc_path = lucas_kanade_rstc(
                    video, start_frame, end_frame, start_points, end_points
                )
                for frame_count, frame_num in enumerate(
                    range(start_frame, end_frame + 1)
                ):
                    target_ann.data[label][frame_num] = list(
                        rstc_path[frame_count, 0, :]
                    )
                rstc_paths[label][start_frame : end_frame + 1, :] = np.squeeze(
                    rstc_path
                )
        self.update()

    def render(
        self, start_frame: int, end_frame: int, out_vid_name: Optional[str] = None
    ) -> None:
        """Render the video with annotations."""
        if out_vid_name is None:
            out_vid_name = str(Path(self.ann.fname).with_suffix(".mp4"))
        assert not os.path.exists(out_vid_name)
        fps = self.ann.video.get_avg_fps()
        codec = "h264"
        writer = FFMpegWriter(fps=fps, codec=codec)
        with writer.saving(self.figure, out_vid_name, dpi=300):
            for idx in range(start_frame, end_frame + 1):
                self._current_idx = idx
                self.update()
                writer.grab_frame()


class VideoAnnotation:
    """
    Manage one point annotation layer in a video.

    Each annotation layer can contain up to 10 labels, which are string representations of digits 0-9.
    Each label is a dictionary mapping a frame number to a 2D location on the video frame.

    Args:
        fname (str, optional): File name of the annotations (.json) file. If it
            doesn't exist, it will be created when the save method is used. If this is a
            video file, `fname` will default to `<video_name>_annotations.json`.
            This can also be a DeepLabCut `.h5` file (either labeled data OR predicted trace).
            Defaults to None.
        vname (str, optional): Name of the video being annotated. Defaults to None.
        name (str, optional): Name of the annotation (something meaningful, e.g., `<muscle_name>_<scorer>`
            such as `brachialis_praneeth`). Defaults to None.
        n_labels (int, optional): Number of labels for annotation. Defaults to 10.
        **kwargs: Additional optional parameters:
            - `palette_name` (str, default='Set2'): Color scheme to use. Defaults to 'Set2' from seaborn.
            - `ax_list` (list, default=[]): If specified, the annotation display will be initialized on these axes.
            Alternatively, use :py:meth:`VideoAnnotation.setup_display` to specify the axis list and colors.
            - `preloaded_json` (dict, optional): The result of `VideoAnnotation._load_json` (in case you prefer
            to pickle the JSON files).

    Methods:
        to_dlc(): Convert from JSON file format into a DeepLabCut DataFrame format, and optionally save the file.
    """

    def __init__(
        self,
        fname: Optional[str] = None,
        vname: Optional[str] = None,
        name: Optional[str] = None,
        n_labels: int = 10,
        **kwargs
    ) -> None:
        self.fname, vname = self._parse_inp(fname, vname, name)

        if self.fname is not None:
            self.fstem = Path(self.fname).stem
        else:
            self.fstem = None

        if name is None:
            if self.fstem is None:
                self.name = "noname"  # default name
            else:
                if "_annotations_" in self.fstem:
                    self.name = self.fstem.split("_annotations_")[-1]
                else:
                    self.name = "noname"
        else:
            assert isinstance(name, str)
            self.name = name

        if utils.is_video(vname):
            self.video = utils.Video(vname)
        else:
            self.video = None

        preloaded_json = kwargs.pop("preloaded_json", None)
        if preloaded_json is None:
            self.data = self.load(n_annotations=n_labels)
        else:
            self.data = preloaded_json

        self._original_palette = utils.get_palette(
            kwargs.pop("palette_name", "Set2"),
            n_colors=1000
        )  # seaborn Set 2
        self.plot_handles = {
            "ax_list_scatter": kwargs.pop("ax_list_scatter", []),
            "ax_list_trace_x": kwargs.pop("ax_list_trace_x", []),
            "ax_list_trace_y": kwargs.pop("ax_list_trace_y", []),
        }
        self._plot_type = "dot" # line or dot
        self.setup_display()

    @classmethod
    def from_multiple_files(
        cls, fname_list: List[str], vname: str, name: str, fname_merged: str, **kwargs
    ) -> VideoAnnotation:
        """Merge annotations from multiple files.
        If multiple files contain an annotation label for the same frame, values from the last file will be kept.
        """
        ann_list: List[VideoAnnotation] = [cls(fname, vname, name, **kwargs) for fname in fname_list]
        assert len({ann.video.name for ann in ann_list}) == 1

        labels = sorted(list({label for ann in ann_list for label in ann.labels}))

        ret = cls(fname=fname_merged, vname=ann_list[-1].video.fname, name=name)
        ret.data = {label: {} for label in labels}
        for label in labels:
            ret.data[label] = functools.reduce(
                lambda x, y: {**x, **y}, [ann.data.get(label, {}) for ann in ann_list]
            )

        return ret

    @staticmethod
    def _parse_inp(fname_inp: Any, vname_inp: Any, name_inp: Any) -> Any:
        if fname_inp is None and vname_inp is None:
            fname, vname = fname_inp, vname_inp  # do nothing, empty annotation
        elif fname_inp is not None and vname_inp is None:
            if utils.is_video(fname_inp):
                vname = fname_inp
                fname = os.path.join(
                    Path(fname_inp).parent, Path(fname_inp).stem + "_annotations.json"
                )
            else:
                fname = fname_inp
                # Try to find the video in the same folder
                vname_potential = os.path.join(
                    Path(fname_inp).parent,
                    utils.removesuffix(Path(fname_inp).stem, "_annotations").split(
                        "_annotations_"
                    )[0]
                    + ".mp4",
                )
                if os.path.exists(vname_potential):
                    vname = vname_potential
                    print(f"Associating video {vname} with the annotation!")
                else:
                    vname = vname_inp
        elif fname_inp is None and vname_inp is not None:
            assert utils.is_video(vname_inp)
            vname = vname_inp
            suffix = "" if name_inp is None else f"_{name_inp}"
            fname = os.path.join(
                Path(vname_inp).parent,
                Path(vname_inp).stem + f"_annotations{suffix}.json",
            )
        elif fname_inp is not None and vname_inp is not None:
            assert utils.is_video(vname_inp)
            fname, vname = fname_inp, vname_inp  # do nothing
        return fname, vname

    def load(self, n_annotations: int = 10, **h5_kwargs) -> dict:
        """Load annotations from a json file, dlc h5 file, or initialize an annotation dictionary if a file doesn't exist."""
        if (self.fname is None) or (not os.path.exists(self.fname)):
            return {str(label): {} for label in range(n_annotations)}

        if Path(self.fname).suffix == ".json":
            return self._load_json(self.fname)

        assert Path(self.fname).suffix == ".h5"
        return self._load_dlc(self.fname, **h5_kwargs)

    @staticmethod
    def _load_json(json_fname: str) -> dict:
        with open(json_fname, "r") as f:
            ret = {}
            for k, v in json.load(f).items():
                ret[k] = {int(frame_num): loc for frame_num, loc in v.items()}
            return ret

    def _load_dlc(self, dlc_fname: str, **kwargs) -> dict:
        if isinstance(dlc_fname, (str, Path)):
            assert os.path.exists(dlc_fname)
            assert Path(dlc_fname).suffix == ".h5"
            df = pd.read_hdf(dlc_fname)
        else:
            assert isinstance(dlc_fname, pd.DataFrame)
            df = dlc_fname
        if isinstance(df.index, pd.MultiIndex):  # labeled data format
            return self._dlc_df_to_annotation_dict(df, **kwargs)
        return self._dlc_trace_to_annotation_dict(
            df, **kwargs
        )  # predicted points trace

    @staticmethod
    def _dlc_df_to_annotation_dict(
        df: pd.DataFrame,
        remove_label_prefix: str = "point",
        img_prefix: str = "img",
        img_suffix: str = ".png",
    ) -> dict:
        """Convert dlc labeled data dataframe to an annotation dictionary"""
        if False in [
            utils.removeprefix(x, remove_label_prefix).isdigit()
            for x in df.columns.levels[1]
        ]:
            label_orig_to_internal = {
                x: str(xcnt) for xcnt, x in enumerate(df.columns.levels[1].tolist())
            }
        else:
            label_orig_to_internal = {
                x: utils.removeprefix(x, remove_label_prefix)
                for x in df.columns.levels[1].tolist()
            }

        frames_str = [
            utils.removesuffix(utils.removeprefix(x, img_prefix), img_suffix)
            for x in df.index.levels[-1]
        ]

        data = {label: {} for label in label_orig_to_internal.values()}
        video_stem = df.index.levels[1].values[0]
        scorer = df.columns.levels[0].values[0]
        for label_orig, label_internal in label_orig_to_internal.items():
            for frame_str in frames_str:
                coord_val = [
                    df.loc[
                        "labeled-data",
                        video_stem,
                        f"{img_prefix}{frame_str}{img_suffix}",
                    ][scorer, label_orig, coord_name]
                    for coord_name in ("x", "y")
                ]
                if np.all(np.isnan(coord_val)):
                    continue
                data[label_internal][int(frame_str)] = coord_val

        return data

    @staticmethod
    def _dlc_trace_to_annotation_dict(
        df: pd.DataFrame, remove_label_prefix: str = "point"
    ) -> dict:
        """Convery dlc labeled trace dataframe (result of analyze_videos) to an annotation dictionary."""
        if False in [
            utils.removeprefix(x, remove_label_prefix).isdigit()
            for x in df.columns.levels[1]
        ]:
            label_orig_to_internal = {
                x: str(xcnt) for xcnt, x in enumerate(df.columns.levels[1].tolist())
            }
        else:
            label_orig_to_internal = {
                x: utils.removeprefix(x, remove_label_prefix)
                for x in df.columns.levels[1].tolist()
            }
        frames = df.index.values
        print(label_orig_to_internal)

        data = {label: {} for label in label_orig_to_internal.values()}
        scorer = df.columns.levels[0].values[0]
        for label_orig, label_internal in label_orig_to_internal.items():
            for frame in frames:
                coord_val = [
                    df.loc[frame][scorer, label_orig, coord_name]
                    for coord_name in ("x", "y")
                ]
                if np.all(np.isnan(coord_val)):
                    continue
                data[label_internal][frame] = coord_val

        return data

    def __len__(self) -> int:
        """Number of annotations"""
        return len(self.data)

    @property
    def n_frames(self) -> int:
        """Number of frames in the video being annotated"""
        if self.video is None:
            return max(self.frames, default=-1) + 1
        return len(self.video)

    @property
    def n_annotations(self) -> int:
        """Number of points being annotated in the video."""
        return len(self)

    @property
    def labels(self) -> List[str]:
        """Labels of the annotations."""
        return list(self.data.keys())
    
    @property
    def palette(self) -> List[Tuple]:
        """Color palette for the annotations."""
        return [self._original_palette[int(label)] for label in self.labels]

    @property
    def frames(self) -> List[int]:
        """Frame numbers in the video that have annotations."""
        ret = list(
            set([frame for label in self.labels for frame in self.get_frames(label)])
        )
        ret.sort()
        return ret

    @property
    def frames_overlapping(self) -> List[int]:
        """List of frames in the video where all the labels are annotated."""
        ret = list(
            functools.reduce(
                set.intersection, [set(self.get_frames(label)) for label in self.labels]
            )
        )
        ret.sort()
        return ret
    
    @property
    def plot_type(self) -> str:
        """Type of plot to use for the annotations."""
        return self._plot_type
    
    @plot_type.setter
    def plot_type(self, plot_type: str) -> None:
        """Set the type of plot to use for the annotations."""
        assert plot_type in ("line", "dot")
        self._plot_type = plot_type
        self.set_plot_type(self._plot_type)

    def get_frames(self, label: str) -> List[int]:
        """Return a list of frames that are annotated with the current label."""
        assert label in self.labels
        return list(self.data[label].keys())

    def save(self, fname: Optional[str] = None) -> None:
        """Save the annotations json file. self.fname should be a valid file path."""
        if fname is None:
            assert self.fname is not None
            fname = self.fname
        # at the moment, saving is only supported for json files through this method
        if Path(fname).suffix != ".json":
            raise ValueError("Supply a json file name.")
        self.sort_data()
        self.remove_empty_labels()
        # cast data due to json dump issues
        data = {
            label: {
                int(frame): [float(x) for x in position]
                for frame, position in label_data.items()
            }
            for label, label_data in self.data.items()
        }
        with open(fname, "w") as f:
            json.dump(data, f, indent=4)
        labels_annotations = {label: len(self.data[label]) for label in self.labels}
        print(f"Saved {fname} with labels-n_annotations \n {labels_annotations}")

    def sort_labels(self) -> None:
        """Sort labels in the data dictionary."""
        self.data = dict(sorted(self.data.items(), key=lambda item: int(item[0])))

    def sort_data(self) -> None:
        """Sort annotations by the frame numbers."""
        self.data = {
            label: dict(sorted(self.data[label].items())) for label in self.labels
        }
    
    def clip_trailing_empty_labels(self) -> None:
        """Remove trailing empty labels from the annotation dictionary."""
        n_labeled_frames = [len(self.data[label]) for label in self.labels]

        def last_nonzero_index(lst):
            for i in range(len(lst) - 1, -1, -1):
                if lst[i] != 0:
                    return i
            return 0  # or raise an exception if needed
        
        last_index = last_nonzero_index(n_labeled_frames)

        self.data = {label: self.data[label] for idx, label in enumerate(self.labels) if idx <= last_index}
    
    def remove_empty_labels(self) -> None:
        """Remove empty labels from the annotation dictionary."""
        self.data = {label: self.data[label] for label in self.labels if len(self.data[label]) > 0}

    def get_values_cv(self, frame_num: int) -> np.ndarray:
        """Return annotations at frame_num in a format for openCV's optical flow algorithms"""
        return np.array(self.get_at_frame(frame_num), dtype=np.float32).reshape(
            (self.n_annotations, 1, 2)
        )

    def _n_digits_in_frame_num(self) -> str:
        """Number of digits to use when constructing a string from the frame number."""
        if self.n_frames is None:
            return "6"
        return str(len(str(self.n_frames)))

    def _frame_num_as_str(self, frame_num: int) -> str:
        """Return the frame umber as a formatted string."""
        return f"{frame_num:0{self._n_digits_in_frame_num()}}"

    def add_at_frame(self, frame_num: int, values: np.ndarray) -> None:
        """Add annotations at a frame, given the annotation values."""
        assert isinstance(frame_num, int)
        values = np.array(values)
        assert values.shape == (self.n_annotations, 2)
        for label, value in zip(self.labels, values):
            self.data[label][frame_num] = list(value)

    def get_at_frame(self, frame_num: int) -> List[List[float]]:
        """Retrieve annotations at a given frame number. If an annotation is not present, nan values will be used."""
        ret = []
        for label in self.labels:
            if frame_num in self.data[label]:
                ret.append(self.data[label][frame_num])
            else:
                ret.append([np.nan, np.nan])
        return ret

    def __getitem__(
        self, key: Union[str, int]
    ) -> Union[Dict[int, List[float]], List[List[float]]]:
        """Easy access to specific annotation, or data from a frame number."""
        if key in self.labels:
            return self.data[key]
        if key in self.frames:
            return self.get_at_frame(key)
        raise ValueError(f"{key} is neither an annotation nor a frame with annotation.")

    def to_dlc(
        self,
        scorer: str = "praneeth",
        output_path: Optional[str] = None,
        file_prefix: Optional[str] = None,
        img_prefix: str = "img",
        img_suffix: str = ".png",
        label_prefix: str = "point",
        save: bool = True,
        internal_to_dlc_labels: Optional[Dict[str, str]] = None,
    ) -> pd.DataFrame:
        """Save annotations in deeplabcut format."""
        if (
            internal_to_dlc_labels is not None
        ):  # label_prefix is ignored if internal_to_dlc_labels are provided
            assert set(internal_to_dlc_labels) == set(self.labels)
            internal_to_dlc_labels = {
                x: internal_to_dlc_labels[x] for x in self.labels
            }  # in case of ordering mishaps
        else:
            internal_to_dlc_labels = {x: f"{label_prefix}{x}" for x in self.labels}

        annotations = self.data

        if output_path is None:
            output_path = Path(self.fname).parent
        output_path = Path(output_path)

        index_length = self._n_digits_in_frame_num()
        img_stems = [
            f"{img_prefix}{x:0{index_length}}{img_suffix}" for x in self.frames
        ]

        row_idx = pd.MultiIndex.from_tuples(
            [("labeled-data", self.video.name, img_stem) for img_stem in img_stems]
        )
        col_idx = pd.MultiIndex.from_product(
            [[scorer], [internal_to_dlc_labels[x] for x in annotations], ["x", "y"]],
            names=["scorer", "bodyparts", "coords"],
        )
        df = pd.DataFrame([], index=row_idx, columns=col_idx)
        for annotation_label_internal, annotation_dict in annotations.items():
            annotation_label_dlc = internal_to_dlc_labels[annotation_label_internal]
            for frame, xy in annotation_dict.items():
                for coord_name, coord_val in zip(("x", "y"), xy):
                    df.loc[
                        "labeled-data",
                        self.video.name,
                        f"{img_prefix}{frame:0{index_length}}{img_suffix}",
                    ][scorer, annotation_label_dlc, coord_name] = coord_val
        df = df.apply(lambda col: pd.to_numeric(col, errors="coerce"))

        if file_prefix is None:
            file_prefix = self.fstem
        elif file_prefix == "dlc":  # usual dlc name
            file_prefix = f"CollectedData_{scorer}"
        else:
            assert isinstance(file_prefix, str)

        if save:
            labeled_data_file_prefix = str(output_path / file_prefix)
            df.to_csv(labeled_data_file_prefix + ".csv")
            df.to_hdf(labeled_data_file_prefix + ".h5", key="df_with_missing", mode="w")
        return df

    def to_trace(self, label: str) -> np.ndarray:
        """Return a 2d numpy array of n_frames x 2.

        Args:
            label (str): Annotation label.

        Returns:
            np.ndarray: 2d numpy array of number of frames x 2.
                xy position values of uannotated frames will be filled with np.nan.
        """
        assert label in self.labels
        ret = np.full([self.n_frames, 2], np.nan)
        for frame_number, frame_xy in self.data[label].items():
            ret[frame_number, :] = frame_xy
        return ret

    def to_traces(self) -> Mapping[str, np.ndarray]:
        """Return annotations as traces (numpy arrays of size n_frames x 2).

        Returns:
            Mapping[str, np.ndarray]: Dictionary mapping labels to 2d numpy arrays.
        """
        return {label: self.to_trace(label) for label in self.labels}

    def to_signal(self, label: str) -> pysampled.Data:
        """Return an annotation as pysampled.Data at the frame rate of the video.

        Args:
            label (str): Annotation label.

        Returns:
            pysampled.Data: Signal sampled at the frame rate of the video being annotated.
        """
        assert self.video is not None
        assert label in self.labels
        return pysampled.Data(self.to_trace(label), sr=self.video.get_avg_fps())

    def to_signals(self) -> Mapping[str, pysampled.Data]:
        """Return annotations as a dictionary of sampled Data signals
        sampled at the frame rate of the video being annotated.

        Returns:
            Mapping[str, pysampled.Data]: Dictionary mapping labels to pysampled.Data.
        """
        return {label: self.to_signal(label) for label in self.labels}

    def add_label(
        self,
        label: Optional[str] = None,
        color: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        if label is None:  # pick the next available label
            label = f"{len(self.labels)}"
        if label in self.labels:
            print(f"Label {label} already exists in layer {self.name}.")
            return
        assert label.isdigit()

        if int(label) > len(self._original_palette):
            self._original_palette = self._original_palette*2
        if color is None:
            color = self._original_palette[int(label)]

        # palette size should mirror the number of labels
        assert len(color) == 3
        assert all([0 <= x <= 1 for x in color])

        self.data[label] = {}
        self.sort_labels()

        self.re_setup_display()

        print(f"Created new label {label} in layer {self.name} with color {color}.")

    def add(self, location: List[float], label: str, frame_number: int) -> None:
        """Add a point annotation (location) of a given label at a frame number."""
        assert len(location) == 2
        self.data[label][frame_number] = list(location)

    def remove(self, label: str, frame_number: int) -> None:
        """Remove a point annotation of a given label at a frame number."""
        assert label in self.labels
        self.data[label].pop(frame_number, None)

    # display management
    def _process_ax_list(
        self, ax_list: Union[None, plt.Axes, List[plt.Axes]], type_: str
    ) -> List[plt.Axes]:
        """Process the list of axes for scatter or trace plots."""
        assert type_ in ("scatter", "trace_x", "trace_y")
        if ax_list is None:
            ax_list = self.plot_handles[f"ax_list_{type_}"]
        if isinstance(ax_list, plt.Axes):
            ax_list = [ax_list]
        assert all([isinstance(ax, plt.Axes) for ax in ax_list])
        self.plot_handles[f"ax_list_{type_}"] = ax_list
        return ax_list

    def setup_display_scatter(
        self, ax_list_scatter: Union[None, plt.Axes, List[plt.Axes]] = None,
    ) -> None:
        """Setup scatter plot display."""
        ax_list_scatter = self._process_ax_list(ax_list_scatter, "scatter")
        palette = self.palette

        # instead of len(self.labels) to keep all 10 points, some of them can be nan
        dummy_xy = [np.nan] * len(palette)
        for ax_cnt, ax in enumerate(ax_list_scatter):
            self.plot_handles[f"labels_in_ax{ax_cnt}"] = ax.scatter(
                dummy_xy, dummy_xy, color=palette, picker=5
            )

    def setup_display_trace(
        self,
        ax_list_trace_x: Union[None, plt.Axes, List[plt.Axes]] = None,
        ax_list_trace_y: Union[None, plt.Axes, List[plt.Axes]] = None,
    ) -> None:
        """Setup trace plot display."""
        ax_list_trace_x = self._process_ax_list(ax_list_trace_x, "trace_x")
        ax_list_trace_y = self._process_ax_list(ax_list_trace_y, "trace_y")

        if self.n_frames > 0 and len(self.frames) / self.n_frames > 0.8:
            plot_type = "-"
        else:
            plot_type = "o"

        x = np.arange(self.n_frames)
        dummy_y = np.full(self.n_frames, np.nan)
        for ax_cnt, (ax_x, ax_y) in enumerate(zip(ax_list_trace_x, ax_list_trace_y)):
            for label, x_color in zip(self.labels, self.palette):
                if ax_x.bbox.bounds == ax_y.bbox.bounds:  # if they are in the same axis
                    y_color = [1 - tc for tc in x_color]
                else:
                    y_color = x_color
                for coord, this_ax, color in zip(
                    ("x", "y"), (ax_x, ax_y), (x_color, y_color)
                ):
                    handle_name = f"trace_in_ax{coord}{ax_cnt}_label{label}"
                    (self.plot_handles[handle_name],) = this_ax.plot(
                        x, dummy_y, plot_type, color=color
                    )

            ax_x.set_xlim(0, self.n_frames)

    def setup_display(
        self,
        ax_list_scatter: Union[None, plt.Axes, List[plt.Axes]] = None,
        ax_list_trace_x: Union[None, plt.Axes, List[plt.Axes]] = None,
        ax_list_trace_y: Union[None, plt.Axes, List[plt.Axes]] = None,
    ) -> None:
        """Setup display for scatter and trace plots."""
        self.setup_display_scatter(ax_list_scatter)
        self.setup_display_trace(ax_list_trace_x, ax_list_trace_y)
        self.set_plot_type(self.plot_type)

    def clear_display(self) -> None:
        """Clear the display."""
        for ax_cnt in range(len(self.plot_handles["ax_list_scatter"])):
            self.plot_handles[f"labels_in_ax{ax_cnt}"].remove()
        for ax_cnt in range(len(self.plot_handles["ax_list_trace_x"])):
            for label in self.labels:
                for ax_type in ("x", "y"):
                    handle_name = f"trace_in_ax{ax_type}{ax_cnt}_label{label}"
                    if handle_name in self.plot_handles:
                        self.plot_handles[handle_name].remove()
        plt.draw()
    
    def re_setup_display(self) -> None:
        """re-establish display elements when adding a label"""
        self.clear_display()
        self.setup_display(
            ax_list_scatter=self.plot_handles["ax_list_scatter"],
            ax_list_trace_x=self.plot_handles["ax_list_trace_x"],
            ax_list_trace_y=self.plot_handles["ax_list_trace_y"],
        )

    def update_display_scatter(self, frame_number: int, draw: bool = False) -> None:
        """Update scatter plot display."""
        for ax_cnt in range(len(self.plot_handles["ax_list_scatter"])):
            n_pts = len(self.palette)
            scatter_offsets = np.full((n_pts, 2), np.nan)
            scatter_offsets[:, :] = self.get_at_frame(frame_number)
            self.plot_handles[f"labels_in_ax{ax_cnt}"].set_offsets(scatter_offsets)
        if draw:
            plt.draw()

    def update_display_trace(
        self, label: Optional[str] = None, draw: bool = False
    ) -> None:
        """Update trace plot display."""
        if label is None:
            label_list = self.labels
        else:
            assert label in self.labels
            label_list = [label]

        for ax_cnt in range(len(self.plot_handles["ax_list_trace_x"])):
            for label in label_list:
                trace = self.to_trace(label)
                for coord_cnt, coord in enumerate(("x", "y")):
                    handle_name = f"trace_in_ax{coord}{ax_cnt}_label{label}"
                    self.plot_handles[handle_name].set_ydata(trace[:, coord_cnt])

        if draw:
            plt.draw()

    def update_display(
        self, frame_number: int, label: Optional[str] = None, draw: bool = False
    ) -> None:
        """Update scatter and trace plot display."""
        self.update_display_scatter(frame_number, draw=False)
        self.update_display_trace(label, draw=False)
        if draw:
            plt.draw()

    # display management - control visibility
    @property
    def _trace_handles(self) -> dict:
        """Dictionary of handle_name - handle for trace handles."""
        return {
            name: handle
            for name, handle in self.plot_handles.items()
            if name.startswith("trace_in_ax")
        }

    @property
    def _label_handles(self) -> dict:
        """Dictionary of handle_name - handle for label handles (image overlay)."""
        return {
            name: handle
            for name, handle in self.plot_handles.items()
            if name.startswith("labels_in_ax")
        }

    @property
    def _trace_or_label_handles(self) -> dict:
        return {**self._trace_handles, **self._label_handles}

    def _set_visibility(self, visibility: bool = True, draw: bool = False) -> None:
        for plot_handle in self._trace_or_label_handles.values():
            plot_handle.set_visible(visibility)
        if draw:
            plt.draw()

    def hide(self, draw: bool = True) -> None:
        """Hide all elements (scatter, traces) in this annotation."""
        self._set_visibility(False, draw)

    def show(self, draw: bool = True) -> None:
        """Show all elements (scatter, traces) in this annotation."""
        self._set_visibility(True, draw)

    def _set_trace_visibility(
        self, label: str, visibility: bool = True, draw: bool = False
    ) -> None:
        for plot_handle_name, plot_handle in self._trace_handles.items():
            if plot_handle_name.endswith(f"_label{label}"):
                plot_handle.set_visible(visibility)
        if draw:
            plt.draw()

    def show_trace(self, label: str, draw: bool = True) -> None:
        """Show trace for a specific label."""
        self._set_trace_visibility(label, True, draw)

    def hide_trace(self, label: str, draw: bool = True) -> None:
        """Hide trace for a specific label."""
        self._set_trace_visibility(label, False, draw)

    def show_one_trace(self, label: str, draw: bool = True) -> None:
        """Show only one trace for a specific label."""
        for this_label in self.labels:
            self._set_trace_visibility(this_label, this_label == label, draw)

    def set_alpha(
        self, alpha: float = 0.4, label: Optional[str] = None, draw: bool = True
    ) -> None:
        """Set the transparency level of all (or one) the traces and labels in this annotation.

        Args:
            alpha (float, optional): alpha value between 0 and 1. Defaults to 0.4.
            label (_type_, optional): Defaults to all labels.
            draw (bool, optional): Update display if True. Defaults to True.
        """
        for handle in self._trace_or_label_handles.values():
            handle.set_alpha(alpha)
        if draw:
            plt.draw()

    def set_plot_type(self, type_: str = "line", draw: bool = True) -> None:
        """Set the plot type for traces.

        Args:
            type_ (str, optional): "line" or "dot". Defaults to "line".
            draw (bool, optional): Update display if True. Defaults to True.
        """
        assert type_ in ("line", "dot")
        for trace_handle in self._trace_handles.values():
            if type_ == "line":
                trace_handle.set_linestyle("-")
                trace_handle.set_marker("None")
            else:
                trace_handle.set_linestyle("None")
                trace_handle.set_marker("o")
        if draw:
            plt.draw()

    def clip_labels(self, start_frame: int, end_frame: int) -> None:
        """Remove annotations outside the clip range. Clip range includes start and end frame."""
        for label in self.labels:
            self.data[label] = {
                k: v
                for k, v in self.data[label].items()
                if start_frame <= k <= end_frame
            }

    def keep_overlapping_continuous_frames(self) -> None:
        """Keep data from consecutive frames that have all labels."""
        x = self.frames_overlapping
        frames_to_keep = sorted(
            set([item for a, b in zip(x, x[1:]) if (b - a) == 1 for item in (a, b)])
        )
        if len(frames_to_keep) == 0:
            print(
                "You're trying to remove all frames! Saving you from yourself by aborting."
            )
            return
        for label in self.labels:
            self.data[label] = {
                k: v for k, v in self.data[label].items() if k in frames_to_keep
            }

    def get_area(
        self, labels: Union[List[str], str], lowpass: Optional[float] = None
    ) -> Union[pysampled.Data, np.ndarray]:
        """Get the area in pixel squared."""

        def PolyArea(x, y):
            return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

        if isinstance(labels, str):
            labels = list(labels)  # e.g. '023' -> ['0', '2', '3']

        for label in labels:
            assert label in self.labels

        if lowpass is None:
            traces = self.to_traces()
        else:
            traces = {
                label: signal.lowpass(lowpass)()
                for label, signal in self.to_signals().items()
            }

        trace_mat = np.asarray([traces[label] for label in labels])
        area_vals = np.array(
            [
                PolyArea(trace_mat[:, xi, 0], trace_mat[:, xi, 1])
                for xi in range(self.n_frames)
            ]
        )

        if self.video is None:  # return np.array
            return area_vals
        return pysampled.Data(area_vals, sr=self.video.get_avg_fps())


class VideoAnnotations(AssetContainer):
    def add(
        self, name: Union[str, VideoAnnotation], fname=None, vname=None, **kwargs
    ) -> VideoAnnotation:
        """Create-and-add"""
        if isinstance(name, VideoAnnotation):
            ann = name
        else:
            assert isinstance(name, str)
            ann = VideoAnnotation(fname, vname, name, **kwargs)
        return super().add(ann)


if __name__ == "__main__":
    import sys
    sys.path.append(r"C:\dev\datanavigator")
    import datanavigator
    video_fname = datanavigator.get_example_video()
    v = VideoPointAnnotator(video_fname, "pn2", n_labels=2)
    for ann in v.annotations:
        ann.add_label("3")
    v.update()
    plt.show(block=True)
